/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


    var bigInt = (function (undefined) {
    "use strict";

    var BASE = 1e7,
        LOG_BASE = 7,
        MAX_INT = 9007199254740992,
        MAX_INT_ARR = smallToArray(MAX_INT),
        LOG_MAX_INT = Math.log(MAX_INT);

    function BigInteger(value, sign) {
        this.value = value;
        this.sign = sign;
        this.isSmall = false;
    }

    function SmallInteger(value) {
        this.value = value;
        this.sign = value < 0;
        this.isSmall = true;
    }

    function isPrecise(n) {
        return -MAX_INT < n && n < MAX_INT;
    }

    function smallToArray(n) { // For performance reasons doesn't reference BASE, need to change this function if BASE changes
        if (n < 1e7)
            return [n];
        if (n < 1e14)
            return [n % 1e7, Math.floor(n / 1e7)];
        return [n % 1e7, Math.floor(n / 1e7) % 1e7, Math.floor(n / 1e14)];
    }

    function arrayToSmall(arr) { // If BASE changes this function may need to change
        trim(arr);
        var length = arr.length;
        if (length < 4 && compareAbs(arr, MAX_INT_ARR) < 0) {
            switch (length) {
                case 0: return 0;
                case 1: return arr[0];
                case 2: return arr[0] + arr[1] * BASE;
                default: return arr[0] + (arr[1] + arr[2] * BASE) * BASE;
            }
        }
        return arr;
    }

    function trim(v) {
        var i = v.length;
        while (v[--i] === 0);
        v.length = i + 1;
    }

    function createArray(length) { // function shamelessly stolen from Yaffle's library https://github.com/Yaffle/BigInteger
        var x = new Array(length);
        var i = -1;
        while (++i < length) {
            x[i] = 0;
        }
        return x;
    }

    function truncate(n) {
        if (n > 0) return Math.floor(n);
        return Math.ceil(n);
    }

    function add(a, b) { // assumes a and b are arrays with a.length >= b.length
        var l_a = a.length,
            l_b = b.length,
            r = new Array(l_a),
            carry = 0,
            base = BASE,
            sum, i;
        for (i = 0; i < l_b; i++) {
            sum = a[i] + b[i] + carry;
            carry = sum >= base ? 1 : 0;
            r[i] = sum - carry * base;
        }
        while (i < l_a) {
            sum = a[i] + carry;
            carry = sum === base ? 1 : 0;
            r[i++] = sum - carry * base;
        }
        if (carry > 0) r.push(carry);
        return r;
    }

    function addAny(a, b) {
        if (a.length >= b.length) return add(a, b);
        return add(b, a);
    }

    function addSmall(a, carry) { // assumes a is array, carry is number with 0 <= carry < MAX_INT
        var l = a.length,
            r = new Array(l),
            base = BASE,
            sum, i;
        for (i = 0; i < l; i++) {
            sum = a[i] - base + carry;
            carry = Math.floor(sum / base);
            r[i] = sum - carry * base;
            carry += 1;
        }
        while (carry > 0) {
            r[i++] = carry % base;
            carry = Math.floor(carry / base);
        }
        return r;
    }

    BigInteger.prototype.add = function (v) {
        var value, n = parseValue(v);
        if (this.sign !== n.sign) {
            return this.subtract(n.negate());
        }
        var a = this.value, b = n.value;
        if (n.isSmall) {
            return new BigInteger(addSmall(a, Math.abs(b)), this.sign);
        }
        return new BigInteger(addAny(a, b), this.sign);
    };
    BigInteger.prototype.plus = BigInteger.prototype.add;

    SmallInteger.prototype.add = function (v) {
        var n = parseValue(v);
        var a = this.value;
        if (a < 0 !== n.sign) {
            return this.subtract(n.negate());
        }
        var b = n.value;
        if (n.isSmall) {
            if (isPrecise(a + b)) return new SmallInteger(a + b);
            b = smallToArray(Math.abs(b));
        }
        return new BigInteger(addSmall(b, Math.abs(a)), a < 0);
    };
    SmallInteger.prototype.plus = SmallInteger.prototype.add;

    function subtract(a, b) { // assumes a and b are arrays with a >= b
        var a_l = a.length,
            b_l = b.length,
            r = new Array(a_l),
            borrow = 0,
            base = BASE,
            i, difference;
        for (i = 0; i < b_l; i++) {
            difference = a[i] - borrow - b[i];
            if (difference < 0) {
                difference += base;
                borrow = 1;
            } else borrow = 0;
            r[i] = difference;
        }
        for (i = b_l; i < a_l; i++) {
            difference = a[i] - borrow;
            if (difference < 0) difference += base;
            else {
                r[i++] = difference;
                break;
            }
            r[i] = difference;
        }
        for (; i < a_l; i++) {
            r[i] = a[i];
        }
        trim(r);
        return r;
    }

    function subtractAny(a, b, sign) {
        var value, isSmall;
        if (compareAbs(a, b) >= 0) {
            value = subtract(a,b);
        } else {
            value = subtract(b, a);
            sign = !sign;
        }
        value = arrayToSmall(value);
        if (typeof value === "number") {
            if (sign) value = -value;
            return new SmallInteger(value);
        }
        return new BigInteger(value, sign);
    }

    function subtractSmall(a, b, sign) { // assumes a is array, b is number with 0 <= b < MAX_INT
        var l = a.length,
            r = new Array(l),
            carry = -b,
            base = BASE,
            i, difference;
        for (i = 0; i < l; i++) {
            difference = a[i] + carry;
            carry = Math.floor(difference / base);
            difference %= base;
            r[i] = difference < 0 ? difference + base : difference;
        }
        r = arrayToSmall(r);
        if (typeof r === "number") {
            if (sign) r = -r;
            return new SmallInteger(r);
        } return new BigInteger(r, sign);
    }

    BigInteger.prototype.subtract = function (v) {
        var n = parseValue(v);
        if (this.sign !== n.sign) {
            return this.add(n.negate());
        }
        var a = this.value, b = n.value;
        if (n.isSmall)
            return subtractSmall(a, Math.abs(b), this.sign);
        return subtractAny(a, b, this.sign);
    };
    BigInteger.prototype.minus = BigInteger.prototype.subtract;

    SmallInteger.prototype.subtract = function (v) {
        var n = parseValue(v);
        var a = this.value;
        if (a < 0 !== n.sign) {
            return this.add(n.negate());
        }
        var b = n.value;
        if (n.isSmall) {
            return new SmallInteger(a - b);
        }
        return subtractSmall(b, Math.abs(a), a >= 0);
    };
    SmallInteger.prototype.minus = SmallInteger.prototype.subtract;

    BigInteger.prototype.negate = function () {
        return new BigInteger(this.value, !this.sign);
    };
    SmallInteger.prototype.negate = function () {
        var sign = this.sign;
        var small = new SmallInteger(-this.value);
        small.sign = !sign;
        return small;
    };

    BigInteger.prototype.abs = function () {
        return new BigInteger(this.value, false);
    };
    SmallInteger.prototype.abs = function () {
        return new SmallInteger(Math.abs(this.value));
    };

    function multiplyLong(a, b) {
        var a_l = a.length,
            b_l = b.length,
            l = a_l + b_l,
            r = createArray(l),
            base = BASE,
            product, carry, i, a_i, b_j;
        for (i = 0; i < a_l; ++i) {
            a_i = a[i];
            for (var j = 0; j < b_l; ++j) {
                b_j = b[j];
                product = a_i * b_j + r[i + j];
                carry = Math.floor(product / base);
                r[i + j] = product - carry * base;
                r[i + j + 1] += carry;
            }
        }
        trim(r);
        return r;
    }

    function multiplySmall(a, b) { // assumes a is array, b is number with |b| < BASE
        var l = a.length,
            r = new Array(l),
            base = BASE,
            carry = 0,
            product, i;
        for (i = 0; i < l; i++) {
            product = a[i] * b + carry;
            carry = Math.floor(product / base);
            r[i] = product - carry * base;
        }
        while (carry > 0) {
            r[i++] = carry % base;
            carry = Math.floor(carry / base);
        }
        return r;
    }

    function shiftLeft(x, n) {
        var r = [];
        while (n-- > 0) r.push(0);
        return r.concat(x);
    }

    function multiplyKaratsuba(x, y) {
        var n = Math.max(x.length, y.length);
        
        if (n <= 30) return multiplyLong(x, y);
        n = Math.ceil(n / 2);

        var b = x.slice(n),
            a = x.slice(0, n),
            d = y.slice(n),
            c = y.slice(0, n);

        var ac = multiplyKaratsuba(a, c),
            bd = multiplyKaratsuba(b, d),
            abcd = multiplyKaratsuba(addAny(a, b), addAny(c, d));

        var product = addAny(addAny(ac, shiftLeft(subtract(subtract(abcd, ac), bd), n)), shiftLeft(bd, 2 * n));
        trim(product);
        return product;
    }

    // The following function is derived from a surface fit of a graph plotting the performance difference
    // between long multiplication and karatsuba multiplication versus the lengths of the two arrays.
    function useKaratsuba(l1, l2) {
        return -0.012 * l1 - 0.012 * l2 + 0.000015 * l1 * l2 > 0;
    }

    BigInteger.prototype.multiply = function (v) {
        var value, n = parseValue(v),
            a = this.value, b = n.value,
            sign = this.sign !== n.sign,
            abs;
        if (n.isSmall) {
            if (b === 0) return CACHE[0];
            if (b === 1) return this;
            if (b === -1) return this.negate();
            abs = Math.abs(b);
            if (abs < BASE) {
                return new BigInteger(multiplySmall(a, abs), sign);
            }
            b = smallToArray(abs);
        }
        if (useKaratsuba(a.length, b.length)) // Karatsuba is only faster for certain array sizes
            return new BigInteger(multiplyKaratsuba(a, b), sign);
        return new BigInteger(multiplyLong(a, b), sign);
    };

    BigInteger.prototype.times = BigInteger.prototype.multiply;

    function multiplySmallAndArray(a, b, sign) { // a >= 0
        if (a < BASE) {
            return new BigInteger(multiplySmall(b, a), sign);
        }
        return new BigInteger(multiplyLong(b, smallToArray(a)), sign);
    }
    SmallInteger.prototype._multiplyBySmall = function (a) {
            if (isPrecise(a.value * this.value)) {
                return new SmallInteger(a.value * this.value);
            }
            return multiplySmallAndArray(Math.abs(a.value), smallToArray(Math.abs(this.value)), this.sign !== a.sign);
    };
    BigInteger.prototype._multiplyBySmall = function (a) {
            if (a.value === 0) return CACHE[0];
            if (a.value === 1) return this;
            if (a.value === -1) return this.negate();
            return multiplySmallAndArray(Math.abs(a.value), this.value, this.sign !== a.sign);
    };
    SmallInteger.prototype.multiply = function (v) {
        return parseValue(v)._multiplyBySmall(this);
    };
    SmallInteger.prototype.times = SmallInteger.prototype.multiply;

    function square(a) {
        var l = a.length,
            r = createArray(l + l),
            base = BASE,
            product, carry, i, a_i, a_j;
        for (i = 0; i < l; i++) {
            a_i = a[i];
            for (var j = 0; j < l; j++) {
                a_j = a[j];
                product = a_i * a_j + r[i + j];
                carry = Math.floor(product / base);
                r[i + j] = product - carry * base;
                r[i + j + 1] += carry;
            }
        }
        trim(r);
        return r;
    }

    BigInteger.prototype.square = function () {
        return new BigInteger(square(this.value), false);
    };

    SmallInteger.prototype.square = function () {
        var value = this.value * this.value;
        if (isPrecise(value)) return new SmallInteger(value);
        return new BigInteger(square(smallToArray(Math.abs(this.value))), false);
    };

    function divMod1(a, b) { // Left over from previous version. Performs faster than divMod2 on smaller input sizes.
        var a_l = a.length,
            b_l = b.length,
            base = BASE,
            result = createArray(b.length),
            divisorMostSignificantDigit = b[b_l - 1],
            // normalization
            lambda = Math.ceil(base / (2 * divisorMostSignificantDigit)),
            remainder = multiplySmall(a, lambda),
            divisor = multiplySmall(b, lambda),
            quotientDigit, shift, carry, borrow, i, l, q;
        if (remainder.length <= a_l) remainder.push(0);
        divisor.push(0);
        divisorMostSignificantDigit = divisor[b_l - 1];
        for (shift = a_l - b_l; shift >= 0; shift--) {
            quotientDigit = base - 1;
            if (remainder[shift + b_l] !== divisorMostSignificantDigit) {
              quotientDigit = Math.floor((remainder[shift + b_l] * base + remainder[shift + b_l - 1]) / divisorMostSignificantDigit);
            }
            // quotientDigit <= base - 1
            carry = 0;
            borrow = 0;
            l = divisor.length;
            for (i = 0; i < l; i++) {
                carry += quotientDigit * divisor[i];
                q = Math.floor(carry / base);
                borrow += remainder[shift + i] - (carry - q * base);
                carry = q;
                if (borrow < 0) {
                    remainder[shift + i] = borrow + base;
                    borrow = -1;
                } else {
                    remainder[shift + i] = borrow;
                    borrow = 0;
                }
            }
            while (borrow !== 0) {
                quotientDigit -= 1;
                carry = 0;
                for (i = 0; i < l; i++) {
                    carry += remainder[shift + i] - base + divisor[i];
                    if (carry < 0) {
                        remainder[shift + i] = carry + base;
                        carry = 0;
                    } else {
                        remainder[shift + i] = carry;
                        carry = 1;
                    }
                }
                borrow += carry;
            }
            result[shift] = quotientDigit;
        }
        // denormalization
        remainder = divModSmall(remainder, lambda)[0];
        return [arrayToSmall(result), arrayToSmall(remainder)];
    }

    function divMod2(a, b) { // Implementation idea shamelessly stolen from Silent Matt's library http://silentmatt.com/biginteger/
        // Performs faster than divMod1 on larger input sizes.
        var a_l = a.length,
            b_l = b.length,
            result = [],
            part = [],
            base = BASE,
            guess, xlen, highx, highy, check;
        while (a_l) {
            part.unshift(a[--a_l]);
            if (compareAbs(part, b) < 0) {
                result.push(0);
                continue;
            }
            xlen = part.length;
            highx = part[xlen - 1] * base + part[xlen - 2];
            highy = b[b_l - 1] * base + b[b_l - 2];
            if (xlen > b_l) {
                highx = (highx + 1) * base;
            }
            guess = Math.ceil(highx / highy);
            do {
                check = multiplySmall(b, guess);
                if (compareAbs(check, part) <= 0) break;
                guess--;
            } while (guess);
            result.push(guess);
            part = subtract(part, check);
        }
        result.reverse();
        return [arrayToSmall(result), arrayToSmall(part)];
    }

    function divModSmall(value, lambda) {
        var length = value.length,
            quotient = createArray(length),
            base = BASE,
            i, q, remainder, divisor;
        remainder = 0;
        for (i = length - 1; i >= 0; --i) {
            divisor = remainder * base + value[i];
            q = truncate(divisor / lambda);
            remainder = divisor - q * lambda;
            quotient[i] = q | 0;
        }
        return [quotient, remainder | 0];
    }

    function divModAny(self, v) {
        var value, n = parseValue(v);
        var a = self.value, b = n.value;
        var quotient;
        if (b === 0) throw new Error("Cannot divide by zero");
        if (self.isSmall) {
            if (n.isSmall) {
                return [new SmallInteger(truncate(a / b)), new SmallInteger(a % b)];
            }
            return [CACHE[0], self];
        }
        if (n.isSmall) {
            if (b === 1) return [self, CACHE[0]];
            if (b == -1) return [self.negate(), CACHE[0]];
            var abs = Math.abs(b);
            if (abs < BASE) {
                value = divModSmall(a, abs);
                quotient = arrayToSmall(value[0]);
                var remainder = value[1];
                if (self.sign) remainder = -remainder;
                if (typeof quotient === "number") {
                    if (self.sign !== n.sign) quotient = -quotient;
                    return [new SmallInteger(quotient), new SmallInteger(remainder)];
                }
                return [new BigInteger(quotient, self.sign !== n.sign), new SmallInteger(remainder)];
            }
            b = smallToArray(abs);
        }
        var comparison = compareAbs(a, b);
        if (comparison === -1) return [CACHE[0], self];
        if (comparison === 0) return [CACHE[self.sign === n.sign ? 1 : -1], CACHE[0]];

        // divMod1 is faster on smaller input sizes
        if (a.length + b.length <= 200)
            value = divMod1(a, b);
        else value = divMod2(a, b);

        quotient = value[0];
        var qSign = self.sign !== n.sign,
            mod = value[1],
            mSign = self.sign;
        if (typeof quotient === "number") {
            if (qSign) quotient = -quotient;
            quotient = new SmallInteger(quotient);
        } else quotient = new BigInteger(quotient, qSign);
        if (typeof mod === "number") {
            if (mSign) mod = -mod;
            mod = new SmallInteger(mod);
        } else mod = new BigInteger(mod, mSign);
        return [quotient, mod];
    }

    BigInteger.prototype.divmod = function (v) {
        var result = divModAny(this, v);
        return {
            quotient: result[0],
            remainder: result[1]
        };
    };
    SmallInteger.prototype.divmod = BigInteger.prototype.divmod;

    BigInteger.prototype.divide = function (v) {
        return divModAny(this, v)[0];
    };
    SmallInteger.prototype.over = SmallInteger.prototype.divide = BigInteger.prototype.over = BigInteger.prototype.divide;

    BigInteger.prototype.mod = function (v) {
        return divModAny(this, v)[1];
    };
    SmallInteger.prototype.remainder = SmallInteger.prototype.mod = BigInteger.prototype.remainder = BigInteger.prototype.mod;

    BigInteger.prototype.pow = function (v) {
        var n = parseValue(v),
            a = this.value,
            b = n.value,
            value, x, y;
        if (b === 0) return CACHE[1];
        if (a === 0) return CACHE[0];
        if (a === 1) return CACHE[1];
        if (a === -1) return n.isEven() ? CACHE[1] : CACHE[-1];
        if (n.sign) {
            return CACHE[0];
        }
        if (!n.isSmall) throw new Error("The exponent " + n.toString() + " is too large.");
        if (this.isSmall) {
            if (isPrecise(value = Math.pow(a, b)))
                return new SmallInteger(truncate(value));
        }
        x = this;
        y = CACHE[1];
        while (true) {
            if (b & 1 === 1) {
                y = y.times(x);
                --b;
            }
            if (b === 0) break;
            b /= 2;
            x = x.square();
        }
        return y;
    };
    SmallInteger.prototype.pow = BigInteger.prototype.pow;

    BigInteger.prototype.modPow = function (exp, mod) {
        exp = parseValue(exp);
        mod = parseValue(mod);
        if (mod.isZero()) throw new Error("Cannot take modPow with modulus 0");
        var r = CACHE[1],
            base = this.mod(mod);
        while (exp.isPositive()) {
            if (base.isZero()) return CACHE[0];
            if (exp.isOdd()) r = r.multiply(base).mod(mod);
            exp = exp.divide(2);
            base = base.square().mod(mod);
        }
        return r;
    };
    SmallInteger.prototype.modPow = BigInteger.prototype.modPow;

    function compareAbs(a, b) {
        if (a.length !== b.length) {
            return a.length > b.length ? 1 : -1;
        }
        for (var i = a.length - 1; i >= 0; i--) {
            if (a[i] !== b[i]) return a[i] > b[i] ? 1 : -1;
        }
        return 0;
    }

    BigInteger.prototype.compareAbs = function (v) {
        var n = parseValue(v),
            a = this.value,
            b = n.value;
        if (n.isSmall) return 1;
        return compareAbs(a, b);
    };
    SmallInteger.prototype.compareAbs = function (v) {
        var n = parseValue(v),
            a = Math.abs(this.value),
            b = n.value;
        if (n.isSmall) {
            b = Math.abs(b);
            return a === b ? 0 : a > b ? 1 : -1;
        }
        return -1;
    };

    BigInteger.prototype.compare = function (v) {
        // See discussion about comparison with Infinity:
        // https://github.com/peterolson/BigInteger.js/issues/61
        if (v === Infinity) {
            return -1;
        }
        if (v === -Infinity) {
            return 1;
        }

        var n = parseValue(v),
            a = this.value,
            b = n.value;
        if (this.sign !== n.sign) {
            return n.sign ? 1 : -1;
        }
        if (n.isSmall) {
            return this.sign ? -1 : 1;
        }
        return compareAbs(a, b) * (this.sign ? -1 : 1);
    };
    BigInteger.prototype.compareTo = BigInteger.prototype.compare;

    SmallInteger.prototype.compare = function (v) {
        if (v === Infinity) {
            return -1;
        }
        if (v === -Infinity) {
            return 1;
        }

        var n = parseValue(v),
            a = this.value,
            b = n.value;
        if (n.isSmall) {
            return a == b ? 0 : a > b ? 1 : -1;
        }
        if (a < 0 !== n.sign) {
            return a < 0 ? -1 : 1;
        }
        return a < 0 ? 1 : -1;
    };
    SmallInteger.prototype.compareTo = SmallInteger.prototype.compare;

    BigInteger.prototype.equals = function (v) {
        return this.compare(v) === 0;
    };
    SmallInteger.prototype.eq = SmallInteger.prototype.equals = BigInteger.prototype.eq = BigInteger.prototype.equals;

    BigInteger.prototype.notEquals = function (v) {
        return this.compare(v) !== 0;
    };
    SmallInteger.prototype.neq = SmallInteger.prototype.notEquals = BigInteger.prototype.neq = BigInteger.prototype.notEquals;

    BigInteger.prototype.greater = function (v) {
        return this.compare(v) > 0;
    };
    SmallInteger.prototype.gt = SmallInteger.prototype.greater = BigInteger.prototype.gt = BigInteger.prototype.greater;

    BigInteger.prototype.lesser = function (v) {
        return this.compare(v) < 0;
    };
    SmallInteger.prototype.lt = SmallInteger.prototype.lesser = BigInteger.prototype.lt = BigInteger.prototype.lesser;

    BigInteger.prototype.greaterOrEquals = function (v) {
        return this.compare(v) >= 0;
    };
    SmallInteger.prototype.geq = SmallInteger.prototype.greaterOrEquals = BigInteger.prototype.geq = BigInteger.prototype.greaterOrEquals;

    BigInteger.prototype.lesserOrEquals = function (v) {
        return this.compare(v) <= 0;
    };
    SmallInteger.prototype.leq = SmallInteger.prototype.lesserOrEquals = BigInteger.prototype.leq = BigInteger.prototype.lesserOrEquals;

    BigInteger.prototype.isEven = function () {
        return (this.value[0] & 1) === 0;
    };
    SmallInteger.prototype.isEven = function () {
        return (this.value & 1) === 0;
    };

    BigInteger.prototype.isOdd = function () {
        return (this.value[0] & 1) === 1;
    };
    SmallInteger.prototype.isOdd = function () {
        return (this.value & 1) === 1;
    };

    BigInteger.prototype.isPositive = function () {
        return !this.sign;
    };
    SmallInteger.prototype.isPositive = function () {
        return this.value > 0;
    };

    BigInteger.prototype.isNegative = function () {
        return this.sign;
    };
    SmallInteger.prototype.isNegative = function () {
        return this.value < 0;
    };

    BigInteger.prototype.isUnit = function () {
        return false;
    };
    SmallInteger.prototype.isUnit = function () {
        return Math.abs(this.value) === 1;
    };

    BigInteger.prototype.isZero = function () {
        return false;
    };
    SmallInteger.prototype.isZero = function () {
        return this.value === 0;
    };
    BigInteger.prototype.isDivisibleBy = function (v) {
        var n = parseValue(v);
        var value = n.value;
        if (value === 0) return false;
        if (value === 1) return true;
        if (value === 2) return this.isEven();
        return this.mod(n).equals(CACHE[0]);
    };
    SmallInteger.prototype.isDivisibleBy = BigInteger.prototype.isDivisibleBy;

    function isBasicPrime(v) {
        var n = v.abs();
        if (n.isUnit()) return false;
        if (n.equals(2) || n.equals(3) || n.equals(5)) return true;
        if (n.isEven() || n.isDivisibleBy(3) || n.isDivisibleBy(5)) return false;
        if (n.lesser(25)) return true;
        // we don't know if it's prime: let the other functions figure it out
    }

    BigInteger.prototype.isPrime = function () {
        var isPrime = isBasicPrime(this);
        if (isPrime !== undefined) return isPrime;
        var n = this.abs(),
            nPrev = n.prev();
        var a = [2, 3, 5, 7, 11, 13, 17, 19],
            b = nPrev,
            d, t, i, x;
        while (b.isEven()) b = b.divide(2);
        for (i = 0; i < a.length; i++) {
            x = bigInt(a[i]).modPow(b, n);
            if (x.equals(CACHE[1]) || x.equals(nPrev)) continue;
            for (t = true, d = b; t && d.lesser(nPrev) ; d = d.multiply(2)) {
                x = x.square().mod(n);
                if (x.equals(nPrev)) t = false;
            }
            if (t) return false;
        }
        return true;
    };
    SmallInteger.prototype.isPrime = BigInteger.prototype.isPrime;

    BigInteger.prototype.isProbablePrime = function (iterations) {
        var isPrime = isBasicPrime(this);
        if (isPrime !== undefined) return isPrime;
        var n = this.abs();
        var t = iterations === undefined ? 5 : iterations;
        // use the Fermat primality test
        for (var i = 0; i < t; i++) {
            var a = bigInt.randBetween(2, n.minus(2));
            if (!a.modPow(n.prev(), n).isUnit()) return false; // definitely composite
        }
        return true; // large chance of being prime
    };
    SmallInteger.prototype.isProbablePrime = BigInteger.prototype.isProbablePrime;

    BigInteger.prototype.next = function () {
        var value = this.value;
        if (this.sign) {
            return subtractSmall(value, 1, this.sign);
        }
        return new BigInteger(addSmall(value, 1), this.sign);
    };
    SmallInteger.prototype.next = function () {
        var value = this.value;
        if (value + 1 < MAX_INT) return new SmallInteger(value + 1);
        return new BigInteger(MAX_INT_ARR, false);
    };

    BigInteger.prototype.prev = function () {
        var value = this.value;
        if (this.sign) {
            return new BigInteger(addSmall(value, 1), true);
        }
        return subtractSmall(value, 1, this.sign);
    };
    SmallInteger.prototype.prev = function () {
        var value = this.value;
        if (value - 1 > -MAX_INT) return new SmallInteger(value - 1);
        return new BigInteger(MAX_INT_ARR, true);
    };

    var powersOfTwo = [1];
    while (powersOfTwo[powersOfTwo.length - 1] <= BASE) powersOfTwo.push(2 * powersOfTwo[powersOfTwo.length - 1]);
    var powers2Length = powersOfTwo.length, highestPower2 = powersOfTwo[powers2Length - 1];

    function shift_isSmall(n) {
        return ((typeof n === "number" || typeof n === "string") && +Math.abs(n) <= BASE) ||
            (n instanceof BigInteger && n.value.length <= 1);
    }

    BigInteger.prototype.shiftLeft = function (n) {
        if (!shift_isSmall(n)) {
            throw new Error(String(n) + " is too large for shifting.");
        }
        n = +n;
        if (n < 0) return this.shiftRight(-n);
        var result = this;
        while (n >= powers2Length) {
            result = result.multiply(highestPower2);
            n -= powers2Length - 1;
        }
        return result.multiply(powersOfTwo[n]);
    };
    SmallInteger.prototype.shiftLeft = BigInteger.prototype.shiftLeft;

    BigInteger.prototype.shiftRight = function (n) {
        var remQuo;
        if (!shift_isSmall(n)) {
            throw new Error(String(n) + " is too large for shifting.");
        }
        n = +n;
        if (n < 0) return this.shiftLeft(-n);
        var result = this;
        while (n >= powers2Length) {
            if (result.isZero()) return result;
            remQuo = divModAny(result, highestPower2);
            result = remQuo[1].isNegative() ? remQuo[0].prev() : remQuo[0];
            n -= powers2Length - 1;
        }
        remQuo = divModAny(result, powersOfTwo[n]);
        return remQuo[1].isNegative() ? remQuo[0].prev() : remQuo[0];
    };
    SmallInteger.prototype.shiftRight = BigInteger.prototype.shiftRight;

    function bitwise(x, y, fn) {
        y = parseValue(y);
        var xSign = x.isNegative(), ySign = y.isNegative();
        var xRem = xSign ? x.not() : x,
            yRem = ySign ? y.not() : y;
        var xBits = [], yBits = [];
        var xStop = false, yStop = false;
        while (!xStop || !yStop) {
            if (xRem.isZero()) { // virtual sign extension for simulating two's complement
                xStop = true;
                xBits.push(xSign ? 1 : 0);
            }
            else if (xSign) xBits.push(xRem.isEven() ? 1 : 0); // two's complement for negative numbers
            else xBits.push(xRem.isEven() ? 0 : 1);

            if (yRem.isZero()) {
                yStop = true;
                yBits.push(ySign ? 1 : 0);
            }
            else if (ySign) yBits.push(yRem.isEven() ? 1 : 0);
            else yBits.push(yRem.isEven() ? 0 : 1);

            xRem = xRem.over(2);
            yRem = yRem.over(2);
        }
        var result = [];
        for (var i = 0; i < xBits.length; i++) result.push(fn(xBits[i], yBits[i]));
        var sum = bigInt(result.pop()).negate().times(bigInt(2).pow(result.length));
        while (result.length) {
            sum = sum.add(bigInt(result.pop()).times(bigInt(2).pow(result.length)));
        }
        return sum;
    }

    BigInteger.prototype.not = function () {
        return this.negate().prev();
    };
    SmallInteger.prototype.not = BigInteger.prototype.not;

    BigInteger.prototype.and = function (n) {
        return bitwise(this, n, function (a, b) { return a & b; });
    };
    SmallInteger.prototype.and = BigInteger.prototype.and;

    BigInteger.prototype.or = function (n) {
        return bitwise(this, n, function (a, b) { return a | b; });
    };
    SmallInteger.prototype.or = BigInteger.prototype.or;

    BigInteger.prototype.xor = function (n) {
        return bitwise(this, n, function (a, b) { return a ^ b; });
    };
    SmallInteger.prototype.xor = BigInteger.prototype.xor;

    var LOBMASK_I = 1 << 30, LOBMASK_BI = (BASE & -BASE) * (BASE & -BASE) | LOBMASK_I;
    function roughLOB(n) { // get lowestOneBit (rough)
        // SmallInteger: return Min(lowestOneBit(n), 1 << 30)
        // BigInteger: return Min(lowestOneBit(n), 1 << 14) [BASE=1e7]
        var v = n.value, x = typeof v === "number" ? v | LOBMASK_I : v[0] + v[1] * BASE | LOBMASK_BI;
        return x & -x;
    }

    function max(a, b) {
        a = parseValue(a);
        b = parseValue(b);
        return a.greater(b) ? a : b;
    }
    function min(a,b) {
        a = parseValue(a);
        b = parseValue(b);
        return a.lesser(b) ? a : b;
    }
    function gcd(a, b) {
        a = parseValue(a).abs();
        b = parseValue(b).abs();
        if (a.equals(b)) return a;
        if (a.isZero()) return b;
        if (b.isZero()) return a;
        var c = CACHE[1], d, t;
        while (a.isEven() && b.isEven()) {
            d = Math.min(roughLOB(a), roughLOB(b));
            a = a.divide(d);
            b = b.divide(d);
            c = c.multiply(d);
        }
        while (a.isEven()) {
            a = a.divide(roughLOB(a));
        }
        do {
            while (b.isEven()) {
                b = b.divide(roughLOB(b));
            }
            if (a.greater(b)) {
                t = b; b = a; a = t;
            }
            b = b.subtract(a);
        } while (!b.isZero());
        return c.isUnit() ? a : a.multiply(c);
    }
    function lcm(a, b) {
        a = parseValue(a).abs();
        b = parseValue(b).abs();
        return a.divide(gcd(a, b)).multiply(b);
    }
    function randBetween(a, b) {
        a = parseValue(a);
        b = parseValue(b);
        var low = min(a, b), high = max(a, b);
        var range = high.subtract(low);
        if (range.isSmall) return low.add(Math.round(Math.random() * range));
        var length = range.value.length - 1;
        var result = [], restricted = true;
        for (var i = length; i >= 0; i--) {
            var top = restricted ? range.value[i] : BASE;
            var digit = truncate(Math.random() * top);
            result.unshift(digit);
            if (digit < top) restricted = false;
        }
        result = arrayToSmall(result);
        return low.add(typeof result === "number" ? new SmallInteger(result) : new BigInteger(result, false));
    }
    var parseBase = function (text, base) {
        var val = CACHE[0], pow = CACHE[1],
            length = text.length;
        if (2 <= base && base <= 36) {
            if (length <= LOG_MAX_INT / Math.log(base)) {
                return new SmallInteger(parseInt(text, base));
            }
        }
        base = parseValue(base);
        var digits = [];
        var i;
        var isNegative = text[0] === "-";
        for (i = isNegative ? 1 : 0; i < text.length; i++) {
            var c = text[i].toLowerCase(),
                charCode = c.charCodeAt(0);
            if (48 <= charCode && charCode <= 57) digits.push(parseValue(c));
            else if (97 <= charCode && charCode <= 122) digits.push(parseValue(c.charCodeAt(0) - 87));
            else if (c === "<") {
                var start = i;
                do { i++; } while (text[i] !== ">");
                digits.push(parseValue(text.slice(start + 1, i)));
            }
            else throw new Error(c + " is not a valid character");
        }
        digits.reverse();
        for (i = 0; i < digits.length; i++) {
            val = val.add(digits[i].times(pow));
            pow = pow.times(base);
        }
        return isNegative ? val.negate() : val;
    };

    function stringify(digit) {
        var v = digit.value;
        if (typeof v === "number") v = [v];
        if (v.length === 1 && v[0] <= 35) {
            return "0123456789abcdefghijklmnopqrstuvwxyz".charAt(v[0]);
        }
        return "<" + v + ">";
    }
    function toBase(n, base) {
        base = bigInt(base);
        if (base.isZero()) {
            if (n.isZero()) return "0";
            throw new Error("Cannot convert nonzero numbers to base 0.");
        }
        if (base.equals(-1)) {
            if (n.isZero()) return "0";
            if (n.isNegative()) return new Array(1 - n).join("10");
            return "1" + new Array(+n).join("01");
        }
        var minusSign = "";
        if (n.isNegative() && base.isPositive()) {
            minusSign = "-";
            n = n.abs();
        }
        if (base.equals(1)) {
            if (n.isZero()) return "0";
            return minusSign + new Array(+n + 1).join(1);
        }
        var out = [];
        var left = n, divmod;
        while (left.isNegative() || left.compareAbs(base) >= 0) {
            divmod = left.divmod(base);
            left = divmod.quotient;
            var digit = divmod.remainder;
            if (digit.isNegative()) {
                digit = base.minus(digit).abs();
                left = left.next();
            }
            out.push(stringify(digit));
        }
        out.push(stringify(left));
        return minusSign + out.reverse().join("");
    }

    BigInteger.prototype.toString = function (radix) {
        if (radix === undefined) radix = 10;
        if (radix !== 10) return toBase(this, radix);
        var v = this.value, l = v.length, str = String(v[--l]), zeros = "0000000", digit;
        while (--l >= 0) {
            digit = String(v[l]);
            str += zeros.slice(digit.length) + digit;
        }
        var sign = this.sign ? "-" : "";
        return sign + str;
    };
    SmallInteger.prototype.toString = function (radix) {
        if (radix === undefined) radix = 10;
        if (radix != 10) return toBase(this, radix);
        return String(this.value);
    };

    BigInteger.prototype.valueOf = function () {
        return +this.toString();
    };
    BigInteger.prototype.toJSNumber = BigInteger.prototype.valueOf;

    SmallInteger.prototype.valueOf = function () {
        return this.value;
    };
    SmallInteger.prototype.toJSNumber = SmallInteger.prototype.valueOf;
    
    function parseStringValue(v) {
            if (isPrecise(+v)) {
                var x = +v;
                if (x === truncate(x))
                    return new SmallInteger(x);
                throw "Invalid integer: " + v;
            }
            var sign = v[0] === "-";
            if (sign) v = v.slice(1);
            var split = v.split(/e/i);
            if (split.length > 2) throw new Error("Invalid integer: " + split.join("e"));
            if (split.length === 2) {
                var exp = split[1];
                if (exp[0] === "+") exp = exp.slice(1);
                exp = +exp;
                if (exp !== truncate(exp) || !isPrecise(exp)) throw new Error("Invalid integer: " + exp + " is not a valid exponent.");
                var text = split[0];
                var decimalPlace = text.indexOf(".");
                if (decimalPlace >= 0) {
                    exp -= text.length - decimalPlace - 1;
                    text = text.slice(0, decimalPlace) + text.slice(decimalPlace + 1);
                }
                if (exp < 0) throw new Error("Cannot include negative exponent part for integers");
                text += (new Array(exp + 1)).join("0");
                v = text;
            }
            var isValid = /^([0-9][0-9]*)$/.test(v);
            if (!isValid) throw new Error("Invalid integer: " + v);
            var r = [], max = v.length, l = LOG_BASE, min = max - l;
            while (max > 0) {
                r.push(+v.slice(min, max));
                min -= l;
                if (min < 0) min = 0;
                max -= l;
            }
            trim(r);
            return new BigInteger(r, sign);
    }
    
    function parseNumberValue(v) {
            if (isPrecise(v)) return new SmallInteger(v);
            return parseStringValue(v.toString());
    }

    function parseValue(v) {
        if (typeof v === "number") {
            return parseNumberValue(v);
        }
        if (typeof v === "string") {
            return parseStringValue(v);
        }
        return v;
    }
    // Pre-define numbers in range [-999,999]
    var CACHE = function (v, radix) {
        if (typeof v === "undefined") return CACHE[0];
        if (typeof radix !== "undefined") return +radix === 10 ? parseValue(v) : parseBase(v, radix);
        return parseValue(v);
    };
    for (var i = 0; i < 1000; i++) {
        CACHE[i] = new SmallInteger(i);
        if (i > 0) CACHE[-i] = new SmallInteger(-i);
    }
    // Backwards compatibility
    CACHE.one = CACHE[1];
    CACHE.zero = CACHE[0];
    CACHE.minusOne = CACHE[-1];
    CACHE.max = max;
    CACHE.min = min;
    CACHE.gcd = gcd;
    CACHE.lcm = lcm;
    CACHE.isInstance = function (x) { return x instanceof BigInteger || x instanceof SmallInteger; };
    CACHE.randBetween = randBetween;
    return CACHE;
})();

// Node.js check
if (typeof module !== "undefined" && module.hasOwnProperty("exports")) {
    module.exports = bigInt;
}





window['jaspy'] = (function () {
    'use strict';

    var $ = {};
    var jaspy = $;

    
        console.info('Jaspy Python Interpreter\nDebugging Mode!');
    

    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var TRACEBACK_ON_EXCEPTION = true;

var CODE_FLAGS = {
    OPTIMIZED: 1 << 0,
    NEWLOCALS: 1 << 1,
    NESTED: 1 << 4,
    GENERATOR: 1 << 5,
    NOFREE: 1 << 6,

    STAR_ARGS: 1 << 2,
    STAR_KWARGS: 1 << 3,

    PYTHON: 1 << 10,
    NATIVE: 1 << 11
};

var BLOCK_TYPES = {
    BASE: 0,
    LOOP: 1,
    EXCEPT: 2,
    FINALLY: 3
};

var CAUSES = {
    RETURN: 0,
    EXCEPTION: 1,
    BREAK: 2,
    CONTINUE: 3,
    YIELD: 4,
    RUN: 5
};

var COMPARE_OPS = {
    LT: 0,
    LE: 1,
    EQ: 2,
    NE: 3,
    GT: 4,
    GE: 5,
    IN: 6,
    NIN: 7,
    IS: 8,
    NIS: 9,
    EXC: 10
};

var COMPARE_SLOTS = [
    '__lt__',
    '__le__',
    '__eq__',
    '__ne__',
    '__gt__',
    '__ge__',
    '__contains__',
    '__contains__'
];

var OPCODES = {
    BEFORE_ASYNC_WITH: 52,
    BINARY_ADD: 23,
    BINARY_AND: 64,
    BINARY_FLOOR_DIVIDE: 26,
    BINARY_LSHIFT: 62,
    BINARY_MATRIX_MULTIPLY: 16,
    BINARY_MODULO: 22,
    BINARY_MULTIPLY: 20,
    BINARY_OR: 66,
    BINARY_POWER: 19,
    BINARY_RSHIFT: 63,
    BINARY_SUBSCR: 25,
    BINARY_SUBTRACT: 24,
    BINARY_TRUE_DIVIDE: 27,
    BINARY_XOR: 65,
    BREAK_LOOP: 80,
    BUILD_LIST: 103,
    BUILD_LIST_UNPACK: 149,
    BUILD_MAP: 105,
    BUILD_MAP_UNPACK: 150,
    BUILD_MAP_UNPACK_WITH_CALL: 151,
    BUILD_SET: 104,
    BUILD_SET_UNPACK: 153,
    BUILD_SLICE: 133,
    BUILD_TUPLE: 102,
    BUILD_TUPLE_UNPACK: 152,
    CALL_FUNCTION: 131,
    CALL_FUNCTION_KW: 141,
    CALL_FUNCTION_VAR: 140,
    CALL_FUNCTION_VAR_KW: 142,
    COMPARE_OP: 107,
    CONTINUE_LOOP: 119,
    DELETE_ATTR: 96,
    DELETE_DEREF: 138,
    DELETE_FAST: 126,
    DELETE_GLOBAL: 98,
    DELETE_NAME: 91,
    DELETE_SUBSCR: 61,
    DUP_TOP: 4,
    DUP_TOP_TWO: 5,
    END_FINALLY: 88,
    EXTENDED_ARG: 144,
    FOR_ITER: 93,
    GET_AITER: 50,
    GET_ANEXT: 51,
    GET_AWAITABLE: 73,
    GET_ITER: 68,
    GET_YIELD_FROM_ITER: 69,
    IMPORT_FROM: 109,
    IMPORT_NAME: 108,
    IMPORT_STAR: 84,
    INPLACE_ADD: 55,
    INPLACE_AND: 77,
    INPLACE_FLOOR_DIVIDE: 28,
    INPLACE_LSHIFT: 75,
    INPLACE_MATRIX_MULTIPLY: 17,
    INPLACE_MODULO: 59,
    INPLACE_MULTIPLY: 57,
    INPLACE_OR: 79,
    INPLACE_POWER: 67,
    INPLACE_RSHIFT: 76,
    INPLACE_SUBTRACT: 56,
    INPLACE_TRUE_DIVIDE: 29,
    INPLACE_XOR: 78,
    JUMP_ABSOLUTE: 113,
    JUMP_FORWARD: 110,
    JUMP_IF_FALSE_OR_POP: 111,
    JUMP_IF_TRUE_OR_POP: 112,
    LIST_APPEND: 145,
    LOAD_ATTR: 106,
    LOAD_BUILD_CLASS: 71,
    LOAD_CLASSDEREF: 148,
    LOAD_CLOSURE: 135,
    LOAD_CONST: 100,
    LOAD_DEREF: 136,
    LOAD_FAST: 124,
    LOAD_GLOBAL: 116,
    LOAD_NAME: 101,
    MAKE_CLOSURE: 134,
    MAKE_FUNCTION: 132,
    MAP_ADD: 147,
    NOP: 9,
    POP_BLOCK: 87,
    POP_EXCEPT: 89,
    POP_JUMP_IF_FALSE: 114,
    POP_JUMP_IF_TRUE: 115,
    POP_TOP: 1,
    PRINT_EXPR: 70,
    RAISE_VARARGS: 130,
    RETURN_VALUE: 83,
    ROT_THREE: 3,
    ROT_TWO: 2,
    SETUP_ASYNC_WITH: 154,
    SETUP_EXCEPT: 121,
    SETUP_FINALLY: 122,
    SETUP_LOOP: 120,
    SETUP_WITH: 143,
    SET_ADD: 146,
    STORE_ATTR: 95,
    STORE_DEREF: 137,
    STORE_FAST: 125,
    STORE_GLOBAL: 97,
    STORE_NAME: 90,
    STORE_SUBSCR: 60,
    UNARY_INVERT: 15,
    UNARY_NEGATIVE: 11,
    UNARY_NOT: 12,
    UNARY_POSITIVE: 10,
    UNPACK_EX: 94,
    UNPACK_SEQUENCE: 92,
    WITH_CLEANUP_FINISH: 82,
    WITH_CLEANUP_START: 81,
    YIELD_FROM: 72,
    YIELD_VALUE: 86
};

var OPCODES_ARGUMENT = 90;

var OPCODES_EXTRA = new Array(255);

OPCODES_EXTRA[OPCODES.UNARY_POSITIVE] = '__pos__';
OPCODES_EXTRA[OPCODES.UNARY_NEGATIVE] = '__neg__';
OPCODES_EXTRA[OPCODES.UNARY_NOT] = '__not__';
OPCODES_EXTRA[OPCODES.UNARY_INVERT] = '__invert__';
OPCODES_EXTRA[OPCODES.GET_ITER] = '__iter__';
OPCODES_EXTRA[OPCODES.GET_YIELD_FROM_ITER] = '__iter__';

OPCODES_EXTRA[OPCODES.BINARY_POWER] = 'pow';
OPCODES_EXTRA[OPCODES.BINARY_MULTIPLY] = 'mul';
OPCODES_EXTRA[OPCODES.BINARY_MATRIX_MULTIPLY] = 'matmul';
OPCODES_EXTRA[OPCODES.BINARY_FLOOR_DIVIDE] = 'floordiv';
OPCODES_EXTRA[OPCODES.BINARY_TRUE_DIVIDE] = 'truediv';
OPCODES_EXTRA[OPCODES.BINARY_MODULO] = 'mod';
OPCODES_EXTRA[OPCODES.BINARY_ADD] = 'add';
OPCODES_EXTRA[OPCODES.BINARY_SUBTRACT] = 'sub';
OPCODES_EXTRA[OPCODES.BINARY_SUBSCR] = 'getitem';
OPCODES_EXTRA[OPCODES.BINARY_LSHIFT] = 'lshift';
OPCODES_EXTRA[OPCODES.BINARY_RSHIFT] = 'rshift';
OPCODES_EXTRA[OPCODES.BINARY_AND] = 'and';
OPCODES_EXTRA[OPCODES.BINARY_XOR] = 'xor';
OPCODES_EXTRA[OPCODES.BINARY_OR] = 'or';

OPCODES_EXTRA[OPCODES.INPLACE_POWER] = '__ipow__';
OPCODES_EXTRA[OPCODES.INPLACE_MULTIPLY] = '__imul__';
OPCODES_EXTRA[OPCODES.INPLACE_MATRIX_MULTIPLY] = '__imatmul__';
OPCODES_EXTRA[OPCODES.INPLACE_FLOOR_DIVIDE] = '__ifloordiv__';
OPCODES_EXTRA[OPCODES.INPLACE_TRUE_DIVIDE] = '__itruediv__';
OPCODES_EXTRA[OPCODES.INPLACE_MODULO] = '__imod__';
OPCODES_EXTRA[OPCODES.INPLACE_ADD] = '__iadd__';
OPCODES_EXTRA[OPCODES.INPLACE_SUBTRACT] = '__isub__';
OPCODES_EXTRA[OPCODES.INPLACE_LSHIFT] = '__ilshift__';
OPCODES_EXTRA[OPCODES.INPLACE_RSHIFT] = '__irshift__';
OPCODES_EXTRA[OPCODES.INPLACE_AND] = '__iand__';
OPCODES_EXTRA[OPCODES.INPLACE_XOR] = '__ixor__';
OPCODES_EXTRA[OPCODES.INPLACE_OR] = '__ior__';
OPCODES_EXTRA[OPCODES.DELETE_SUBSCR] = '__delitem__';

OPCODES_EXTRA[OPCODES.STORE_SUBSCR] = '__setitem__';

OPCODES_EXTRA[OPCODES.SETUP_LOOP] = BLOCK_TYPES.LOOP;
OPCODES_EXTRA[OPCODES.SETUP_EXCEPT] = BLOCK_TYPES.EXCEPT;
OPCODES_EXTRA[OPCODES.SETUP_FINALLY] = BLOCK_TYPES.FINALLY;

    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function error(message) {
    throw new Error('[FATAL ERROR] ' + (message || 'fatal interpreter error'));
}

function raise(exc_type, exc_value, exc_tb) {
    error('exception occurred before jaspy has been fully initialized');
}

function assert(condition, message) {
    if (!condition) {
        error(message || 'assertion failed');
    }
}

function assign(target, source) {
    var name, value;
    for (name in source) {
        if (source.hasOwnProperty(name)) {
            target[name] = source[name];
        }
    }
}


function make_class(attributes, superclass) {
    var constructor = attributes.constructor;
    if (superclass) {
        constructor.prototype = Object.create(superclass.prototype);
    }
    assign(constructor.prototype, attributes);
    constructor.superclass = superclass;
    constructor.extend = function (attributes) {
        return make_class(attributes, constructor);
    };
    return constructor;
}

var Class = make_class({constructor: function () {}});


$.error = error;
$.raise = raise;
$.assert = assert;
$.assign = assign;

$.Class = Class;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var Signature = Class.extend({
    constructor: function (argnames, poscount, var_args, var_kwargs) {
        this.argnames = argnames || [];
        this.poscount = poscount || 0;
        this.var_args = var_args || false;
        this.var_kwargs = var_kwargs || false;
    },

    parse_args: function (args, kwargs, defaults, namespace) {
        args = args || [];
        kwargs = kwargs || {};

        var result, index, name, length;
        if (!namespace) {
            result = [];
        }

        for (index = 0; index < this.poscount; index++) {
            name = this.argnames[index];
            if (args[index]) {
                if (name in kwargs) {
                    raise(TypeError, 'multiple values for positional argument \'' + name + '\'');
                }
                if (namespace) {
                    namespace[name] = args[index];
                } else {
                    result.push(args[index]);
                }
            } else if (name in kwargs) {
                if (namespace) {
                    namespace[name] = kwargs[name];
                } else {
                    result.push(kwargs[name]);
                }
                delete kwargs[name];
            } else if (defaults && name in defaults) {
                if (namespace) {
                    namespace[name] = defaults[name];
                } else {
                    result.push(defaults[name]);
                }
            } else {
                raise(TypeError, 'missing positional argument \'' + name + '\'');
            }
        }
        if (this.var_args) {
            if (namespace) {
                namespace[this.argnames[index]] = args.slice(this.poscount);
            } else {
                result.push(args.slice(this.poscount));
            }
            index++;
        } else if (index < args.length) {
            raise(TypeError, 'too many positional arguments');
        }

        length = this.argnames.length;
        if (this.var_kwargs) {
            length--;
        }
        for (; index < length; index++) {
            name = this.argnames[index];
            if (name in kwargs) {
                if (namespace) {
                    namespace[name] = kwargs[name];
                } else {
                    result.push(kwargs[name]);
                }
                delete kwargs[name];
            } else if (defaults && name in defaults) {
                if (namespace) {
                    namespace[name] = defaults[name];
                } else {
                    result.push(defaults[name]);
                }
            } else {
                raise(TypeError, 'missing keyword argument \'' + name + '\'');
            }
        }
        if (this.var_kwargs) {
            if (namespace) {
                namespace[this.argnames[index]] = kwargs;
            } else {
                result.push(kwargs);
            }
        } else {
            for (name in kwargs) {
                if (kwargs.hasOwnProperty(name)) {
                    raise(TypeError, 'unknown keyword argument \'' + name + '\'');
                }
            }
        }

        return result;
    },

    toString: function () {
        var argnames;

        if (!this.var_args && !this.var_kwargs) {
            return this.argnames.join(',');
        }
        argnames = new Array(this.argnames);
        if (this.var_args) {
            argnames[this.poscount] = '*' + argnames[this.poscount];
        }
        if (this.var_kwargs) {
            argnames[argnames.length - 1] = '**' + argnames[argnames.length - 1];
        }

        return argnames.join(',');
    }
});

Signature.from_spec = function (spec) {
    var index, name, star_args, star_kwargs;
    var poscount = 0;
    var argnames = [];

    for (index = 0; index < spec.length; index++) {
        name = spec[index];
        if (name.indexOf('**') == 0) {
            if (index != spec.length - 1) {
                raise(TypeError, 'invalid native signature specification');
            }
            name = name.substring(2);
            star_kwargs = true;
        } else if (name.indexOf('*') == 0) {
            if (star_args) {
                raise(TypeError, 'invalid native signature specification');
            }
            name = name.substring(1);
            star_args = true;
        } else if (!star_args) {
            poscount++;
        }
        argnames.push(name);
    }

    return new Signature(argnames, poscount, star_args, star_kwargs);
};

Signature.from_python = function (varnames, argcount, kwargcount, flags) {
    var index;
    var argnames = [];
    var var_args = (flags & CODE_FLAGS.STAR_ARGS) != 0;
    var var_kwargs = (flags & CODE_FLAGS.STAR_KWARGS) != 0;

    for (index = 0; index < argcount; index++) {
        argnames.push(varnames[index]);
    }
    if (var_args) {
        argnames.push(varnames[argcount + kwargcount]);
    }
    for (; index < argcount + kwargcount; index++) {
        argnames.push(varnames[index]);
    }
    if (var_kwargs) {
        argnames.push(varnames[argcount + kwargcount + 1]);
    }

    return new Signature(argnames, argcount, var_args, var_kwargs);
};


var Code = Class.extend({
    constructor: function (signature, options) {
        this.signature = signature;

        options = options || {};

        this.name = options.name || '<unknown>';
        this.filename = options.filename || '<unknown>';

        this.flags = options.flags || 0;
    },

    parse_args: function (args, kwargs, defaults, namespace) {
        return this.signature.parse_args(args, kwargs, defaults, namespace);
    }
});


var PythonCode = Code.extend({
    constructor: function (bytecode, options) {
        this.bytecode = bytecode;

        options = options || {};
        options.flags = (options.flags || 0) | CODE_FLAGS.PYTHON;
        options.name = options.name || '<module>';
        options.filename = options.filename || '<string>';

        this.names = options.names || [];
        this.varnames = options.varnames || [];
        this.freevars = options.freevars || [];
        this.cellvars = options.cellvars || [];

        this.argcount = options.argcount || 0;
        this.kwargcount = options.kwargcount || 0;

        Code.call(this, Signature.from_python(this.varnames, this.argcount, this.kwargcount, options.flags), options);

        this.constants = options.constants || [];

        this.firstline = options.firstline || 1;
        this.lnotab = options.lnotab || '';

        this.instructions = disassemble(this);
    },

    get_line_number: function (position) {
        if (position < 0) {
            return 0;
        }

        var index, offset_increment, line_increment;
        var address = this.instructions[position].start;
        var line_number = this.firstline;
        var offset = 0;

        for (index = 0; index < this.lnotab.length; index++) {
            offset_increment = this.lnotab.charCodeAt(index++);
            line_increment = this.lnotab.charCodeAt(index);
            offset += offset_increment;
            if (offset > address) {
                break;
            }
            line_number += line_increment;
        }

        return line_number;
    }
});

var NativeCode = Code.extend({
    constructor: function (func, options, spec) {
        this.func = func;

        options = options || {};
        options.flags = (options.flags || 0) | CODE_FLAGS.NATIVE;
        options.name = options.name || func.name;
        options.filename = options.filename || '<native>';

        Code.call(this, Signature.from_spec(spec), options);

        this.simple = func.length == this.signature.argnames.length;
    }
});


$.Signature = Signature;

$.Code = Code;
$.PythonCode = PythonCode;
$.NativeCode = NativeCode;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var modules = {};
var pending = {};

function get_module(name) {
    if (name in modules) {
        return modules[name];
    } else {
        raise(ImportError, 'no module named \'' + name + '\'');
    }
}

function get_namespace(name) {
    return get_module(name).dict;
}

function register_module(name, module) {
    modules[name] = module;
    module.dict['__name__'] = pack_str(name);
}

function unregister_module(name) {
    delete modules[name];
}


var Module = Class.extend({
    constructor: function (name, depends) {
        this.name = name;
        this.depends = depends || [];
        this.dict = {};
        if (this.name) {
            register_module(this.name, this);
        }
        this.wrapper = null;
    }
});


var PythonModule = Module.extend({
    constructor: function (name, code, depends) {
        Module.call(this, name, depends);
        this.code = code;
        this.frame = null;
    }
});


var NativeModule = Module.extend({
    constructor: function (name, func, depends) {
        Module.call(this, name, depends);
        this.func = func;
        if (func) {
            func.apply(null, [jaspy, this].concat(this.depends.map(get_namespace)));
        }
    },

    $def: function (name, func, signature, options) {
        options = options || {};
        signature = signature || [];
        options.module = this.name;
        options.name = name;
        options.qualname = name;
        options.simple = func.length == signature.length;
        this.dict[name] = $def(func, signature, options);
        return this.dict[name];
    },

    $set: function (name, value) {
        this.dict[name] = value;
        return this.dict[name];
    },

    $class: function (name, bases, mcs) {
        this.dict[name] = PyType.native(name, bases, null, mcs);
        return this.dict[name];
    }
});

function module(name, initializer, depends) {
    if (initializer instanceof PyCode) {
        initializer = initializer.code;
    }
    if (typeof initializer == 'function') {
        return new NativeModule(name, initializer, depends);
    } else if (initializer instanceof PythonCode) {
        return new PythonModule(name, initializer, depends);
    } else {
        throw new Error('invalid type of code or function');
    }
}


$.get_module = get_module;
$.get_namespace = get_namespace;

$.module = module;

$.modules = modules;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var object_id_counter = 0;


var PyObject = Class.extend({
    constructor: function (cls, dict) {
        this.cls = cls;
        this.identity = null;
        this.dict = dict === undefined ? {} : dict;
    },

    get_id: function () {
        if (this.identity === null) {
            this.identity = object_id_counter++;
        }
        return this.identity;
    },

    get_address: function () {
        return ('0000000000000' + this.get_id().toString(16)).substr(-13);
    },

    call: function (name, args, kwargs) {
        var method = this.cls.lookup(name);
        if (method) {
            if (isinstance(method, py_classmethod)) {
                return call(method.func, [this.cls].concat(args || []), kwargs);
            } else if (isinstance(method, py_staticmethod)) {
                return call(method.func, args, kwargs);
            } else if (name == '__new__') {
                return call(method, [this.cls].concat(args || []), kwargs);
            }
            return call(method, [this].concat(args || []), kwargs);
        } else {
            vm.return_value = null;
            vm.last_exception = METHOD_NOT_FOUND;
            return false;
        }
    },

    setattr: function (name, value) {
        if (!this.dict) {
            raise(TypeError, 'object does not support attribute access');
        }
        if (name instanceof PyStr) {
            name = name.value;
        } else if (typeof name != 'string') {
            raise(TypeError, 'native attribute name must be a string');
        }
        this.dict[name] = value;
    },

    getattr: function (name) {
        if (!this.dict) {
            raise(TypeError, 'object does not support attribute access');
        }
        if (name instanceof PyStr) {
            name = name.value;
        } else if (typeof name != 'string') {
            raise(TypeError, 'native attribute name must be a string');
        }
        return this.dict[name]
    },

    is: function (other) {
        return this === other;
    },

    repr: function () {
        return '<' + this.cls.name + ' object at 0x' + this.get_address() + '>';
    },


    bool: function () {
        return true;
    },

    number: function () {
        raise(TypeError, 'unable to convert object to native number');
    },

    string: function () {
        raise(TypeError, 'unable to convert object to native string');
    }
});

PyObject.prototype.toString = PyObject.prototype.repr;


function isinstance(object, cls) {
    var index;
    for (index = 0; index < object.cls.mro.length; index++) {
        if (object.cls.mro[index] === cls) {
            return true;
        }
    }
    return false;
}


$.PyObject = PyObject;

$.isinstance = isinstance;

    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


function get_mro(cls) {
    return cls.mro;
}

function linearize(cls) {
    var pending = cls.bases.map(get_mro), result = [cls];
    var index, head, good;
    while (pending.length != 0) {
        for (index = 0; index < pending.length; index++) {
            head = pending[index][0];
            good = true;
            pending.forEach(function (base_mro) {
                base_mro.slice(1, base_mro.length).forEach(function (base_cls) {
                    good &= base_cls != head;
                })
            });
            if (good) {
                result.push(head);
                break;
            }
        }
        if (!good) {
            raise(TypeError, 'unable to linearize class hierarchy');
        }
        for (index = 0; index < pending.length; index++) {
            pending[index] = pending[index].filter(function (base_cls) {
                return base_cls != head;
            })
        }
        pending = pending.filter(function (base_mro) {
            return base_mro.length > 0;
        });
    }
    return result;
}


var PyType = PyObject.extend({
    constructor: function (name, bases, attributes, mcs) {
        var index, native;
        PyObject.call(this, mcs || py_type, attributes || {});
        this.name = name;
        this.bases = bases || [py_object];
        this.mro = linearize(this);
        this.native = null;
        for (index = 0; index < this.mro.length; index++) {
            native = this.mro[index].native;
            if (native === py_object) {
                continue;
            }
            if (this.native && this.native !== native && native) {
                raise(TypeError, 'invalid native type hierarchy');
            }
            this.native = native;
        }
        this.native = this.native || py_object;
    },

    lookup: function (name) {
        var index, value;
        for (index = 0; index < this.mro.length; index++) {
            value = this.mro[index].getattr(name);
            if (value) {
                return value;
            }
        }
    },

    define: function (name, item) {
        return this.dict[name] = item;
    },

    $def: function (name, func, signature, options) {
        options = options || {};
        options.name = options.name || name;
        options.qualname = options.qualname || (this.name + '.' + options.name);
        return this.define(name, $def(func, ['self'].concat(signature || []), options));
    },

    $def_property: function (name, getter, setter) {
        var options = {name: name, qualname: this.name + '.' + name};
        if (getter) {
            getter = $def(getter, ['self'], options);
        }
        if (setter) {
            setter = $def(setter, ['self', 'value'], options);
        }
        return this.define(name, new_property(getter, setter));
    },

    $def_classmethod: function (name, func, signature, options) {
        options = options || {};
        options.name = options.name || name;
        options.qualname = options.qualname || (this.name + '.' + options.name);
        return this.define(name, $def(func, ['cls'].concat(signature || []), options));
    },

    $def_alias: function (name, alias) {
        return this.define(alias, this.lookup(name));
    },

    call_classmethod: function (name, args, kwargs) {
        var method = this.lookup(name);
        if (method) {
            return call(method, [this].concat(args || []), kwargs);
        } else {
            vm.return_value = null;
            vm.last_exception = METHOD_NOT_FOUND;
            return false;
        }
    },

    call_staticmethod: function (name, args, kwargs) {
        var method = this.lookup(name);
        if (method) {
            return call(method, args, kwargs);
        } else {
            vm.return_value = null;
            vm.last_exception = METHOD_NOT_FOUND;
            return false;
        }
    },

    check: function (object) {
        if (!isinstance(object, this)) {
            raise(TypeError, 'native type check failed');
        }
    },

    check_subclass: function (cls) {
        if (!issubclass(cls, this)) {
            raise(TypeError, 'native subclass check failed');
        }
    },

    make: function (dict) {
        return new PyObject(this, dict)
    }
});



PyType.native = function (name, bases, attributes, mcs) {
    var type = new PyType(name, bases, attributes, mcs);
    type.native = type;
    return type;
};

function $class(name, bases, attributes, mcs) {
    return new PyType.native(name, bases, attributes, mcs);
}


function issubclass(cls, superclass) {
    var index;
    if (!(cls instanceof PyType)) {
        raise(TypeError, 'first argument must be a class')
    }
    for (index = 0; index < cls.mro.length; index++) {
        if (cls.mro[index] === superclass) {
            return true;
        }
    }
    return false;
}


$.PyType = PyType;

$.$class = $class;

$.issubclass = issubclass;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyDict = PyObject.extend({
    constructor: function (namespace, cls) {
        PyObject.call(this, cls || py_dict);
        this.table = namespace || {};
        if (!(this.table instanceof Object)) {
            raise(TypeError, 'invalid type of native dict initializer');
        }
    },

    get: function (str_key, fallback) {
        if (str_key instanceof PyStr) {
            str_key = str_key.value;
        } else if (typeof str_key != 'string') {
            raise(TypeError, 'invalid native dict key type');
        }
        return this.table[str_key] || fallback || null;
    },

    set: function (str_key, value) {
        if (typeof str_key == 'string') {
            str_key = pack_str(str_key);
        } else if (!(str_key instanceof PyStr)) {
            raise(TypeError, 'invalid native dict key type');
        }
        this.table[str_key] = value;
    },

    pop: function (str_key) {
        var value;
        if (str_key instanceof PyStr) {
            str_key = str_key.value;
        } else if (typeof str_key != 'string') {
            raise(TypeError, 'invalid native dict key type');
        }
        value = this.table[str_key];
        delete this.table[str_key];
        return value;
    }
});


$.PyDict = PyDict;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var PyInt = PyObject.extend({
    constructor: function (value, cls) {
        PyObject.call(this, cls || py_int);
        try {
            this.value = bigInt(value);
        } catch (error) {
            raise(TypeError, 'invalid type of native int initializer');
        }
    },

    toString: function () {
        return this.value.toString();
    },

    valueOf: function () {
        return this.number();
    },

    bool: function () {
        return this.value.neq(0);
    },

    is: function (other) {
        if (other instanceof PyInt) {
            return this.value.eq(other.value);
        }
        return false;
    },

    number: function () {
        var number = this.value.toJSNumber();
        if (number == Infinity || number == -Infinity) {
            raise(OverflowError, 'int too large to convert to float');
        }
        return number;
    },

    float: function () {
        return new PyFloat(this.number());
    },

    abs: function () {
        return new PyInt(this.value.abs());
    },

    pos: function () {
        return this;
    },

    neg: function () {
        return new PyInt(this.value.negate());
    },

    invert: function () {
        return new PyInt(this.value.not());
    },

    add: function (other) {
        return new PyInt(this.value.add(other.value));
    },

    sub: function (other) {
        return new PyInt(this.value.subtract(other.value));
    },

    pow: function (other) {
        if (other < 0) {
            return pack_float(Math.pow(this.number(), other.number()));
        }
        return new PyInt(this.value.pow(other.value));
    },

    mul: function (other) {
        return new PyInt(this.value.multiply(other.value));
    },

    floordiv: function (other) {
        return new PyInt(this.value.divide(other.value));
    },

    truediv: function (other) {
        return pack_float(this.number() / other.number());
    },

    mod: function (other) {
        if (!this.value.sign && other.value.sign) {
            return new PyInt(this.value.mod(other.value).substract(other.value));
        } else if (this.value.sign && !other.value.sign) {
            return new PyInt(other.value.substract(this.value.mod(other.value)));
        }
        return new PyInt(this.value.mod(other.value));
    },

    lshift: function (other) {
        return new PyInt(this.value.shiftLeft(other.value));
    },


    rshift: function (other) {
        return new PyInt(this.value.shiftRight(other.value));
    },

    and: function (other) {
        return new PyInt(this.value.and(other.value));
    },

    xor: function (other) {
        return new PyInt(this.value.xor(other.value));
    },

    or: function (other) {
        return new PyInt(this.value.or(other.value));
    },

    lt: function (other) {
        return this.value.lt(other.value);
    },

    le: function (other) {
        return this.value.leq(other.value);
    },

    eq: function (other) {
        return this.value.eq(other.value);
    },

    ne: function (other) {
        return this.value.neq(other.value);
    },

    gt: function (other) {
        return this.value.gt(other.value);
    },

    ge: function (other) {
        return this.value.geq(other.value);
    }
});


PyInt.parse = function (string, base) {
    if (base instanceof PyInt) {
        return new PyInt(bigInt(string, base.value));
    }
    raise(TypeError, 'invalid type of integer base');
};


$.PyInt = PyInt;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyFloat = PyObject.extend({
    constructor: function (value, cls) {
        if (!(typeof value == 'number')) {
            raise(TypeError, 'invalid type of native float initializer');
        }
        PyObject.call(this, cls || py_float);
        this.value = value;
    },

    number: function () {
        return this.value;
    }
});


$.PyFloat = PyFloat;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyStr = PyObject.extend({
    constructor: function (value, cls) {
        PyObject.call(this, cls || py_str);
        this.value = value;
    },

    repr: function () {
        return '\'' + this.value + '\''
    },

    split: function (sep) {
        return this.value.split(sep);
    },

    encode: function (encoding) {
        var encoder, result;
        if (!TextEncoder) {
            // Polyfill: https://github.com/inexorabletash/text-encoding
            raise(RuntimeError, 'browser does not support encoding, please use a polyfill');
        }
        try {
            encoder = new TextEncoder(encoding || 'utf-8');
        } catch (error) {
            raise(LookupError, 'unknown encoding: ' + encoding);
        }
        try {
            result = encoder.encode(this.value);
        } catch (error) {
            console.log(error);
            raise(UnicodeEncodeError, 'unable to decode bytes object, data is not valid');
        }
        return result;
    },

    toString: function () {
        return this.value;
    }
});


$.PyStr = PyStr;

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var SINGLE_QUOTES_CODE = 39;
var DOUBLE_QUOTES_CODE = 34;


var PyBytes = PyObject.extend({
    constructor: function (value, cls) {
        if (!(value instanceof Uint8Array)) {
            raise(TypeError, 'invalid type of native bytes initializer');
        }
        PyObject.call(this, cls || py_bytes);
        this.array = value;
    },

    bool: function () {
        return this.array.length != 0;
    },

    get: function (offset) {
        return this.array[offset];
    },

    repr: function () {
        var offset, char;
        var chars = new Array(this.array.length);
        for (offset = 0; offset < chars.length; offset++) {
            char = this.array[offset];
            if (char == SINGLE_QUOTES_CODE) {
                chars[offset] = '\\\'';
            } else if (char >= 32 && char <= 126) {
                chars[offset] = String.fromCharCode(char);
            } else {
                chars[offset] = '\\x' + ('0' + char.toString(16)).substr(-2);
            }
        }
        return 'b\'' + chars.join('') + '\'';
    },

    decode: function (encoding) {
        var decoder, result;
        if (!TextDecoder) {
            // Polyfill: https://github.com/inexorabletash/text-encoding
            raise(RuntimeError, 'browser does not support decoding, please enable the polyfill');
        }
        try {
            decoder = new TextDecoder(encoding || 'utf-8', {fatal: true});
        } catch (error) {
            raise(LookupError, 'unknown encoding: ' + encoding);
        }
        try {
            result = decoder.decode(this.array);
        } catch (error) {
            raise(UnicodeDecodeError, 'unable to decode bytes object, data is not valid');
        }
        return result;
    }
});

PyBytes.prototype.toString = PyBytes.prototype.repr;


$.PyBytes = PyBytes;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyTuple = PyObject.extend({
    constructor: function (array, cls) {
        if (!(array instanceof Array)) {
            raise(TypeError, 'invalid type of native tuple initializer');
        }
        PyObject.call(this, cls || py_tuple);
        this.array = array;
    },

    get: function (index) {
        return this.array[index];
    },

    len: function () {
        return this.array.length;
    }
});


$.PyTuple = PyTuple;

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyList = PyObject.extend({
    constructor: function (initializer, size, cls) {
        if (!(initializer instanceof Array)) {
            raise(TypeError, 'invalid type of list initializer');
        }
        PyObject.call(this, cls || py_list);
        this.value = new Array(4);
        if (initializer) {
            this.size = initializer.length;
        } else {
            this.size = size || 0;
        }
        this.grow();
        if (initializer) {
            for (var index = 0; index < initializer.length; index++) {
                this.value[index] = initializer[index];
            }
        }
    },

    check: function (index) {
        if (index < 0) {
            index = this.size - index;
        }
        if (index < 0 || index > this.size - 1) {
            raise(IndexError, 'index out of range');
        }
        return index;
    },

    grow: function () {
        while (this.value.length <= this.size) {
            var length = this.value.length * 2;
            while (length <= this.size) {
                length *= 2;
            }
            this.value.length = length;
        }
    },

    shrink: function () {
        if (this.value.length > 4 && this.value.length / 4 >= this.size) {
            var length = this.value.length / 2;
            while (length / 4 >= this.size && length > 4) {
                length /= 2;
            }
            this.value.length = length;
        }
    },

    get: function (index) {
        index = this.check(index);
        return this.value[index] || None;
    },

    set: function (index, item) {
        index = this.check(index);
        return this.value[index] = item;
    },

    append: function (item) {
        this.size++;
        this.grow();
        this.value[this.size - 1] = item;
        return item;
    },

    pop: function (index) {
        index = this.check(index);
        this.size--;
        if (index == null) {
            index = this.size;
        }
        var item = this.value[index];
        for (; index < this.size; index++) {
            this.value[index] = this.value[index + 1];
        }
        this.value[index] = null;
        this.shrink();
        return item;
    },

    clear: function () {
        this.value = new Array(4);
        this.size = 0;
    },

    slice: function (start, stop, step) {
        var index, list = new PyList();
        if (start == undefined) {
            start = 0;
        } else if (start < 0) {
            start = this.size + start;
        }
        if (stop == undefined) {
            stop = this.size;
        } else if (stop < 0) {
            stop = this.size + stop;
        }
        step = step || 1;
        if (step > 0) {
            if (start < 0) {
                start = 0;
            }
            if (stop > this.size) {
                stop = this.size;
            }
            for (index = start; index < stop; index += step) {
                list.append(this.value[index]);
            }
        } else if (step < 0) {
            if (start >= this.size) {
                start = this.size - 1;
            }
            if (stop < 0) {
                stop = 0;
            }
            for (index = start; index > stop; index += step) {
                list.append(this.value[index]);
            }
        } else {
            raise(ValueError, 'slice step cannot be zero')
        }
        return list;
    },

    concat: function (list_or_array) {
        var list, index, size;
        if (list_or_array instanceof PyList) {
            size = list_or_array.size;
            list_or_array = list_or_array.value;
        } else if (list_or_array instanceof Array) {
            size = list_or_array.length;
        } else {
            raise(TypeError, 'invalid type of concatenation object');
        }
        list = new PyList(null, this.size + size);
        for (index = 0; index < this.size; index++) {
            list.value[index] = this.value[index];
        }
        for (index = 0; index < size; index++) {
            list.value[index + this.size] = list_or_array[index];
        }
        return list;
    },

    copy: function () {
        return this.concat([]);
    }
});


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyCode = PyObject.extend({
    constructor: function (code, cls) {
        if (!(code instanceof Code)) {
            raise(TypeError, 'invalid type of native code initializer');
        }
        PyObject.call(this, cls || py_code);
        this.code = code;
    }
});


function new_code(value) {
    return new PyCode(value);
}


$.PyCode = PyCode;

$.new_code = new_code;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyCell = PyObject.extend({
    constructor: function (object) {
        PyObject.call(this, py_cell);
        this.object = object;
    },

    set: function (object) {
        this.object = object;
    },

    get: function () {
        return this.object;
    }
});


$.PyCell = PyCell;

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyFrame = PyObject.extend({
    constructor: function (frame) {
        if (!(frame instanceof Frame)) {
            raise(TypeError, 'invalid type of native frame initializer');
        }
        PyObject.call(this, py_frame);
        this.frame = frame;
    }
});


$.PyFrame = PyFrame;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyJSObject = PyObject.extend({
    constructor: function (object, cls) {
        if (!(object instanceof Object)) {
            raise(TypeError, 'invalid type of native object initializer');
        }
        PyObject.call(this, cls || py_js_object);
        this.object = object;
    },

    primitive: function () {
        return this.object;
    }
});


var PyJSArray = PyObject.extend({
    constructor: function (array, cls) {
        if (!(array instanceof Array)) {
            raise(TypeError, 'invalid type of native array initializer');
        }
        PyObject.call(this, cls || py_js_array);
        this.array = array;
    },

    primitive: function () {
        return this.array;
    }
});


var PyJSFunction = PyObject.extend({
    constructor: function (func, cls) {
        if (typeof func != 'function') {
            raise(TypeError, 'invalid type of native function initializer');
        }
        PyObject.call(this, cls || py_js_function);
        this.func = func;
    },

    primitive: function () {
        return this.func;
    }
});


$.PyJSObject = PyJSObject;
$.PyJSArray = PyJSArray;
$.PyJSFunction = PyJSFunction;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyFunction = PyObject.extend({
    constructor: function (name, code, options, cls) {
        PyObject.call(this, cls || py_function);

        this.name = name;
        this.code = code;

        this.qualname = options.qualname || this.name;
        this.doc = options.doc || '';
        this.module = options.module || 'builtins';
        this.defaults = options.defaults || null;
        this.closure = options.closure || null;
        this.globals = options.globals || null;
    }
});


function $def(func, signature, options) {
    options = options || {};
    var name = options.name || '<unknown>';
    var code = new NativeCode(func, options, signature);
    return new PyFunction(name, code, options);
}

$.PyFunction = PyFunction;

$.$def = $def;

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyGenerator = PyObject.extend({
    constructor: function (code, frame) {
        PyObject.call(this, py_generator);
        this.code = code;
        this.frame = frame;
    }
});


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function new_property(getter, setter) {
    return new PyObject(py_property, {
        'fget': getter || None,
        'fset': setter || None
    });
}


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PySlice = PyObject.extend({
    constructor: function (start, stop, step) {
        PyObject.call(this, py_slice);
        this.start = start;
        this.stop = stop;
        this.step = step || None;
    }
});


function new_slice(start, stop, step) {
    return new PySlice(start, stop, step);
}


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyException = PyObject.extend({
    constructor: function (args, cls) {
        PyObject.call(this, cls);
        this.args = args;
    }
});


function make_exception(cls, message) {
    var exc_value = new PyObject(cls, {});
    exc_value.dict['args'] = pack_tuple([pack_str(message)]);
    return exc_value;
}




function format_exception(exc_value) {
    var string = [];
    if (exc_value.traceback) {
        string.push(format_traceback(exc_value.traceback));
    }
    if (exc_value.getattr('args') instanceof PyTuple && exc_value.getattr('args').array[0] instanceof PyStr) {
        string.push(exc_value.cls.name + ': ' + exc_value.getattr('args').array[0]);
    } else {
        string.push(exc_value.cls.name);
    }
    return string.join('\n');
}

function print_exception(exc_value) {
    console.error(format_exception(exc_value));
}

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var PyTraceback = PyObject.extend({
    constructor: function (frame, position, line, next) {
        PyObject.call(this, py_traceback);
        this.frame = frame;
        this.position = position;
        this.line = line;
        this.next = next || None;
    }
});


function format_traceback(traceback) {
    var string = ['Traceback (most recent call last):'];
    while (traceback && traceback != None) {
        string.push('    File \'' + traceback.frame.code.filename + '\', line ' + traceback.line + ', in ' + traceback.frame.code.name);
        traceback = traceback.next;
    }
    return string.join('\n');
}

function print_traceback(traceback) {
    console.error(format_traceback(traceback));
}


$.PyTraceback = PyTraceback;



    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var py_object = PyType.native('object', []);
var py_type = PyType.native('type', [py_object]);

py_object.cls = py_type;
py_type.cls = py_type;

var py_dict = PyType.native('dict');

var py_int = PyType.native('int');
var py_bool = PyType.native('bool', [py_int]);

var py_float = PyType.native('float');

var py_str = PyType.native('str');
var py_bytes = PyType.native('bytes');

var py_tuple = PyType.native('tuple');

var py_code = PyType.native('code');

var py_list = PyType.native('list');

var py_cell = PyType.native('cell');
var py_frame = PyType.native('frame');

var py_js_object = PyType.native('JSObject');
var py_js_array = PyType.native('JSArray');
var py_js_function = PyType.native('JSFunction');

var py_traceback = PyType.native('traceback');

var py_function = PyType.native('function');
var py_method = PyType.native('method');
var py_generator = PyType.native('generator');

var py_set = PyType.native('set');
var py_frozenset = PyType.native('frozenset');

var py_classmethod = PyType.native('classmethod');
var py_staticmethod = PyType.native('staticmethod');

var py_module = PyType.native('ModuleType');

var py_property = PyType.native('property');

var py_slice = PyType.native('slice');

var None = new PyObject(PyType.native('NoneType'));
var NotImplemented = new PyObject(PyType.native('NotImplemented'));
var Ellipsis = new PyObject(PyType.native('Ellipsis'));

var False = new PyInt(0, py_bool);
var True = new PyInt(1, py_bool);

var BaseException = PyType.native('BaseException');
var Exception = new PyType('Exception', [BaseException]);
var ValueError = new PyType('ValueError', [Exception]);
var ArithmeticError = new PyType('ArithmeticError', [Exception]);
var LookupError = new PyType('LookupError', [Exception]);
var RuntimeError = new PyType('RuntimeError', [Exception]);
var BufferError = new PyType('BufferError', [Exception]);
var AssertionError = new PyType('AssertionError', [Exception]);
var AttributeError = new PyType('AttributeError', [Exception]);
var EOFError = new PyType('EOFError', [Exception]);
var FloatingPointError = new PyType('FloatingPointError', [ArithmeticError]);
var GeneratorExit = new PyType('GeneratorExit', [BaseException]);
var ImportError = new PyType('ImportError', [Exception]);
var IndexError = new PyType('IndexError', [LookupError]);
var KeyError = new PyType('KeyError', [Exception]);
var KeyboardInterrupt = new PyType('KeyboardInterrupt', [BaseException]);
var MemoryError = new PyType('MemoryError', [Exception]);
var NameError = new PyType('NameError', [Exception]);
var NotImplementedError = new PyType('NotImplementedError', [RuntimeError]);
var OSError = new PyType('OSError', [Exception]);
var OverflowError = new PyType('OverflowError', [Exception]);
var RecursionError = new PyType('RecursionError', [RuntimeError]);
var ReferenceError = new PyType('ReferenceError', [Exception]);
var StopIteration = new PyType('StopIteration', [Exception]);
var SyntaxError = new PyType('SyntaxError', [Exception]);
var IndentationError = new PyType('IndentationError', [SyntaxError]);
var TabError = new PyType('TabError', [IndentationError]);
var SystemError = new PyType('SystemError', [Exception]);
var SystemExit = new PyType('SystemExit', [BaseException]);
var TypeError = new PyType('TypeError', [Exception]);
var UnboundLocalError = new PyType('UnboundLocalError', [NameError]);
var UnicodeError = new PyType('UnicodeError', [ValueError]);
var UnicodeEncodeError = new PyType('UnicodeEncodeError', [UnicodeError]);
var UnicodeDecodeError = new PyType('UnicodeDecodeError', [UnicodeError]);
var UnicodeTranslateError = new PyType('UnicodeTranslateError', [UnicodeError]);
var ZeroDivisionError = new PyType('ZeroDivisionError', [ArithmeticError]);
var EnvironmentError = OSError;
var IOError = OSError;

var BlockingIOError = new PyType('BlockingIOError', [OSError]);
var ChildProcessError = new PyType('ChildProcessError', [OSError]);
var BrokenPipeError = new PyType('BrokenPipeError', [OSError]);
var ConnectionError = new PyType('ConnectionError', [OSError]);
var ConnectionAbortedError = new PyType('ConnectionAbortedError', [ConnectionError]);
var ConnectionRefusedError = new PyType('ConnectionRefusedError', [ConnectionError]);
var ConnectionResetError = new PyType('ConnectionResetError', [ConnectionError]);
var FileExistsError = new PyType('FileExistsError', [OSError]);
var FileNotFoundError = new PyType('FileNotFoundError', [OSError]);
var InterruptedError = new PyType('InterruptedError', [OSError]);
var IsADirectoryError = new PyType('IsADirectoryError', [OSError]);
var NotADirectoryError = new PyType('NotADirectoryError', [OSError]);
var PermissionError = new PyType('PermissionError', [OSError]);
var ProcessLookupError = new PyType('ProcessLookupError', [OSError]);
var TimeoutError = new PyType('TimeoutError', [OSError]);

var MethodNotFoundError = new PyType('MethodNotFoundError', [TypeError]);
var METHOD_NOT_FOUND = {
    exc_type: MethodNotFoundError,
    exc_value: make_exception(MethodNotFoundError, 'method not found'),
    exc_tb: None
};

var UnpackError = new PyType('UnpackError', [TypeError]);
var PackError = new PyType('PackError', [TypeError]);

var JSError = new PyType('JSError', [Exception]);

var ZERO = new PyInt(0);
var ONE = new PyInt(1);

var EMPTY_TUPLE = new PyTuple([]);


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function pack_int(value) {
    return new PyInt(value);
}

function pack_bool(boolean) {
    return boolean ? True : False;
}

function pack_float(value) {
    return new PyFloat(value);
}

function pack_str(value) {
    return new PyStr(value);
}

function pack_bytes(array) {
    return new PyBytes(array);
}

function pack_tuple(array) {
    return new PyTuple(array);
}

function pack_object(object) {
    return new PyJSObject(object);
}

function pack_array(array) {
    return new PyJSArray(array);
}

function pack_function(func) {
    return new PyJSFunction(func);
}

function pack_error(error) {
    return new PyException(pack_tuple([pack_str(error.name), pack_str(error.message)]), JSError);
}

function pack(object) {
    if (object == undefined) {
        return None;
    }
    if (object instanceof PyObject) {
        return object;
    } else if (typeof object == 'string') {
        return pack_str(object);
    } else if (typeof object == 'number') {
        if (Number.isInteger(object)) {
            return pack_int(object);
        } else {
            return pack_float(object);
        }
    } else if (typeof object == 'boolean') {
        return pack_bool(object);
    } else if (typeof object == 'function') {
        return pack_function(object);
    } else if (object instanceof Array) {
        return pack_array(object);
    } else if (object instanceof Object) {
        return pack_object(object);
    } else {
        raise(TypeError, 'unable to pack native object');
    }
}

function unpack_int(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyInt)) {
        raise(TypeError, 'unable to unpack integer from object');
    }
    return object.number();
}

function unpack_bool(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object.cls === py_bool)) {
        raise(TypeError, 'unable to unpack bool from object');
    }
    return object === True;
}

function unpack_number(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyInt) && !(object instanceof PyFloat)) {
        raise(TypeError, 'unable to unpack number from object');
    }
    return object.number();
}

function unpack_str(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyStr)) {
        raise(TypeError, 'unable to unpack string from object');
    }
    return object.value;
}

function unpack_bytes(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyBytes)) {
        raise(TypeError, 'unable to unpack bytes from object');
    }
    return object.array;
}

function unpack_tuple(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyTuple)) {
        raise(TypeError, 'unable to unpack tuple from object');
    }
    return object.array;
}

function unpack_object(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyJSObject)) {
        raise(TypeError, 'unable to unpack js object');
    }
    return object.object;
}

function unpack_array(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyJSArray)) {
        raise(TypeError, 'unable to unpack js array');
    }
    return object.array;
}

function unpack_function(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyJSFunction)) {
        raise(TypeError, 'unable to unpack js function');
    }
    return object.func;
}

function unpack(object, fallback) {
    if ((object === None || !object) && fallback) {
        return fallback;
    }
    if (!(object instanceof PyObject)) {
        raise(TypeError, 'object to unpack is not a python object');
    }
    if (object.cls === py_bool) {
        return unpack_bool(object);
    } else if (object === None) {
        return null;
    } else if (object instanceof PyInt) {
        return unpack_int(object);
    } else if (object instanceof PyFloat) {
        return unpack_float(object);
    } else if (object instanceof PyStr) {
        return unpack_str(object);
    } else if (object instanceof PyBytes) {
        return unpack_bytes(object);
    } else if (object instanceof PyTuple) {
        return unpack_tuple(object);
    } else if (object instanceof PyJSObject) {
        return unpack_object(object);
    } else if (object instanceof PyJSArray) {
        return unpack_array(object);
    } else if (object instanceof PyJSFunction) {
        return unpack_function(object);
    } else {
        raise(TypeError, 'unable to unpack native value from object');
    }
}



$.pack_int = pack_int;
$.pack_bool = pack_bool;
$.pack_float = pack_float;
$.pack_str = pack_str;
$.pack_bytes = pack_bytes;
$.pack_tuple = pack_tuple;
$.pack_object = pack_object;
$.pack_array = pack_array;
$.pack_function = pack_function;
$.pack_error = pack_error;

$.unpack_int = unpack_int;
$.unpack_bool = unpack_bool;
$.unpack_number = unpack_number;
$.unpack_str = unpack_str;
$.unpack_bytes = unpack_bytes;
$.unpack_tuple = unpack_tuple;
$.unpack_object = unpack_object;
$.unpack_array = unpack_array;
$.unpack_function = unpack_function;

$.pack = pack;
$.unpack = unpack;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_object.$def('__new__', function (cls, args, kwargs) {
    if (!(cls instanceof PyType)) {
        raise(TypeError, 'object.__new__(X): X is not a type object');
    }
    if (cls.native !== py_object) {
        raise(TypeError, 'object.__new__() is not safe, use ' + cls.native.name + '.__new__()');
    }
    return new PyObject(cls, {});
}, ['*args', '**kwargs']);

py_object.$def('__getattribute__', function (self, name, state, frame) {
    var value;
    switch (state) {
        case 0:
            name = unpack_str(name);
            value = self.dict ? self.getattr(name) : null;
            if (!value) {
                value = self.cls.lookup(name);
                if (value) {
                    if (value.call('__get__', [self, self.cls])) {
                        return 1;
                    }
                } else {
                    raise(AttributeError, '\'' + self.cls.name + '\' object has no attribute \'' + name + '\'');
                }
            } else {
                return value;
            }
        case 1:
            if (except(MethodNotFoundError)) {
                return value;
            } else if (vm.return_value) {
                return vm.return_value
            } else {
                return null;
            }
    }
}, ['name']);

py_object.$def('__setattr__', function (self, name, item, state, frame) {
    var descriptor;
    switch (state) {
        case 0:
            descriptor = self.cls.lookup(name);
            if (descriptor && descriptor.cls.lookup('__set__')) {
                if (descriptor.call('__set__', [self, item])) {
                    return 1;
                }
            } else {
                self.setattr(name, item);
                return null;
            }
        case 1:
            return null;
    }
}, ['name', 'item']);

py_object.$def('__repr__', function (self) {
    var module = self.cls.getattr('__module__');
    if (module instanceof PyStr) {
        return pack_str('<' + module.value + '.' + self.cls.name + ' object at 0x' + self.get_address() + '>');
    } else {
        return pack_str('<' + self.cls.name + ' object at 0x' + self.get_address() + '>');
    }
});

py_object.$def('__str__', function (self, state, frame) {
    switch (state) {
        case 0:
            if (self.call('__repr__')) {
                return 1;
            }
        case 1:
            return vm.return_value;
    }
});

py_object.$def('__hash__', function (self) {
    return pack_str('object: ' + self.get_address());
});

py_object.$def('__eq__', function (self, other) {
    return self === other ? True : False;
}, ['other']);

py_object.$def_property('__class__', function (self) {
    return self.cls;
}, function (self, value) {
    if (!(value instanceof PyType) || value.native != py_object) {
        raise(TypeError, 'invalid type of \'value\' argument');
    }
    if (self instanceof PyType || self.cls.native != py_object) {
        raise(TypeError, 'object does not support class assignment');
    }
    self.cls = value;
});

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_type.$def_classmethod('__prepare__', function (mcs, bases) {
    return new PyDict();
}, ['bases']);

py_type.$def('__new__', function (mcs, name, bases, attributes) {
    if (!(mcs instanceof PyType)) {
        raise(TypeError, 'invalid type of \'mcs\' argument');
    }
    if (!(attributes instanceof PyDict)) {
        raise(TypeError, 'invalid type of \'attributes\' argument');
    }
    return new PyType(unpack_str(name), unpack_tuple(bases), attributes.table, mcs);
}, ['name', 'bases', 'attributes']);

py_type.$def('__call__', function (cls, args, kwargs, state, frame) {
    switch (state) {
        case 0:
            if (cls.call_classmethod('__new__', args, kwargs)) {
                return 1;
            }
        case 1:
            if (!vm.return_value) {
                return null;
            }
            frame.instance = vm.return_value;
            if (vm.return_value.cls.lookup('__init__')) {
                if (vm.return_value.call('__init__', args, kwargs)) {
                    return 2;
                }
            }
        case 2:
            if (vm.return_value) {
                return frame.instance;
            }
    }
}, ['*args', '**kwargs']);

py_type.$def('__str__', function (cls) {
    var module = cls.getattr('__module__');
    if (!(cls instanceof PyType)) {
        raise(TypeError, 'invalid type of \'cls\' argument');
    }
    if (module instanceof PyStr) {
        return pack_str('<class \'' + unpack_str(module) + '.' + cls.name + '\'>');
    } else {
        return pack_str('<class \'' + cls.name + '\'>');
    }
});

py_type.$def_property('__name__', function (cls) {
    py_type.check(cls);
    return pack_str(cls.name);
}, function (cls, value) {
    py_type.check(cls);
    cls.name = unpack_str(value);
});

py_type.$def_property('__mro__', function (cls) {
    py_type.check(cls);
    return pack_tuple(cls.mro);
});
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_dict.$def('__setitem__', function (self, key, value, state, frame) {
    if (!(self instanceof PyDict)) {
        raise(TypeError, 'invalid type of \'self\' argument');
    }
    switch (state) {
        case 0:
            if (key.cls === py_str) {
                vm.return_value = key;
            } else if (key.call('__hash__')) {
                return 1;
            }
        case 1:
            if (!vm.return_value) {
                return null;
            }
            if (vm.return_value instanceof PyStr) {
                self.set(vm.return_value, value);
            } else if (vm.return_value instanceof PyInt) {
                self.set(pack_str(vm.return_value.value.toString()), value);
            } else {
                raise(TypeError, 'invalid result type of key hash');
            }
            return None;
    }
}, ['key', 'value']);

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_int.$def('__new__', function (cls, initializer, base, state, frame) {
    switch (state) {
        case 0:
            if (!(issubclass(cls, py_int))) {
                raise(TypeError, 'class is not an subclass of int');
            }
            if (initializer instanceof PyFloat) {
                return pack_int(Math.floor(initializer.value), cls);
            }
            if (initializer instanceof PyInt) {
                if (initializer.cls == cls) {
                    return initializer;
                } else {
                    return pack_int(initializer.value, cls);
                }
            }
            if (initializer instanceof PyStr) {
                return PyInt.parse(initializer.value, base);
            }
            if (initializer.call('__int__')) {
                return 1;
            }
        case 1:
            if (except(MethodNotFoundError)) {
                raise(TypeError, 'invalid type of int initializer');
            } else if (vm.return_value) {
                return vm.return_value;
            }
            break;
    }
}, ['initializer', 'base'], {defaults: {initializer: pack_int(0), base: pack_int(10)}});

py_int.$def('__str__', function (self) {
    return pack_str(self.toString());
});

py_int.$def_alias('__str__', '__repr__');

py_int.$def('__bool__', function (self) {
    return pack_bool(self.ne(False));
});

py_int.$def('__neg__', function (self) {
    return self.neg();
});

py_int.$def('__pos__', function (self) {
    return self;
});

py_int.$def('__lt__', function (self, other) {
    return pack_bool(self.lt(other));
}, ['other']);

py_int.$def('__le__', function (self, other) {
    return pack_bool(self.le(other));
}, ['other']);

py_int.$def('__eq__', function (self, other) {
    return pack_bool(self.eq(other));
}, ['other']);

py_int.$def('__ne__', function (self, other) {
    return pack_bool(self.ne(other));
}, ['other']);

py_int.$def('__gt__', function (self, other) {
    return pack_bool(self.gt(other));
}, ['other']);

py_int.$def('__ge__', function (self, other) {
    return pack_bool(self.ge(other));
}, ['other']);

py_int.$def('__pow__', function (self, other) {
    return self.pow(other);
}, ['other']);
py_int.$def_alias('__pow__', '__ipow__');
py_int.$def_alias('__pow__', '__rpow__');

py_int.$def('__mul__', function (self, other) {
    return self.mul(other);
}, ['other']);
py_int.$def_alias('__mul__', '__imul__');
py_int.$def_alias('__mul__', '__rmul__');

py_int.$def('__floordiv__', function (self, other) {
    return self.floordiv(other);
}, ['other']);
py_int.$def_alias('__floordiv__', '__ifloordiv__');
py_int.$def_alias('__floordiv__', '__rfloordiv__');

py_int.$def('__truediv__', function (self, other) {
    return self.truediv(other);
}, ['other']);
py_int.$def_alias('__truediv__', '__itruediv__');
py_int.$def_alias('__truediv__', '__rtruediv__');

py_int.$def('__mod__', function (self, other) {
    return self.mod(other);
}, ['other']);
py_int.$def_alias('__mod__', '__imod__');
py_int.$def_alias('__mod__', '__rmod__');

py_int.$def('__add__', function (self, other) {
    return self.add(other);
}, ['other']);
py_int.$def_alias('__add__', '__iadd__');
py_int.$def_alias('__add__', '__radd__');

py_int.$def('__sub__', function (self, other) {
    return self.sub(other);
}, ['other']);
py_int.$def_alias('__sub__', '__isub__');
py_int.$def_alias('__sub__', '__rsub__');

py_int.$def('__lshift__', function (self, other) {
    return self.lshift(other);
}, ['other']);
py_int.$def_alias('__lshift__', '__ilshift__');
py_int.$def_alias('__lshift__', '__rlshift__');

py_int.$def('__rshift__', function (self, other) {
    return self.rshift(other);
}, ['other']);
py_int.$def_alias('__rshift__', '__irshift__');
py_int.$def_alias('__rshift__', '__rrshift__');

py_int.$def('__and__', function (self, other) {
    return self.and(other);
}, ['other']);
py_int.$def_alias('__and__', '__iand__');
py_int.$def_alias('__and__', '__rand__');

py_int.$def('__xor__', function (self, other) {
    return self.xor(other);
}, ['other']);
py_int.$def_alias('__xor__', '__ixor__');
py_int.$def_alias('__xor__', '__rxor__');

py_int.$def('__or__', function (self, other) {
    return self.or(other);
}, ['other']);
py_int.$def_alias('__or__', '__ior__');
py_int.$def_alias('__or__', '__ror__');

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var TRUE_STR = pack_str('True');
var FALSE_STR = pack_str('False');

py_bool.$def('__new__', function (cls, initializer, state, frame) {
    switch (state) {
        case 0:
            // FIXME: subclassing bool is not allowed
            if (!(issubclass(cls, py_bool))) {
                raise(TypeError, 'class is not an subclass of bool');
            }
            if (initializer.call('__bool__')) {
                return 1;
            }
        case 1:
            if (except(MethodNotFoundError)) {
                if (initializer.call('__len__')) {
                    return 2;
                }
            } else {
                // FIXME: create object of class cls
                return new vm.return_value;
            }
        case 2:
            if (except(MethodNotFoundError)) {
                return True;
            } else if (vm.return_value) {
                // FIXME: create object of class cls
                return vm.return_value.ne(False) ? True : False;
            }
    }
}, ['initializer'], {defaults: {initializer: False}});

py_bool.$def('__str__', function (self) {
    return self.ne(False) ? TRUE_STR : FALSE_STR;
});

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_float.$def('__new__', function (cls, initializer, state, frame) {
    switch (state) {
        case 0:
            if (!(subclass(cls, py_float))) {
                raise(TypeError, 'class is not an subclass of float');
            }
            if (initializer instanceof PyInt || initializer instanceof PyFloat) {
                if (initializer.cls == cls) {
                    return initializer;
                } else {
                    return pack_int(initializer.value, cls);
                }
            }
            if (initializer instanceof PyStr) {
                return pack_float(parseFloat(initializer.value));
            }
            if (initializer.call('__float__')) {
                return 1;
            }
        case 1:
            if (except(MethodNotFoundError)) {
                raise(TypeError, 'invalid type of int initializer');
            } else if (vm.return_value) {
                return vm.return_value;
            }
            break;
    }
}, ['initializer'], {defaults: {initializer: pack_float(0)}});

py_float.$def('__str__', function (self) {
    return pack_str(unpack_number(self).toString());
});

py_float.$def('__neg__', function (self) {
    return pack_float(-unpack_number(self));
});

py_float.$def('__pos__', function (self) {
    return self;
});

py_float.$def('__lt__', function (self, other) {
    return unpack_number(self) < unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__le__', function (self, other) {
    return unpack_number(self) <= unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__eq__', function (self, other) {
    return unpack_number(self) == unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__ne__', function (self, other) {
    return unpack_number(self) != unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__gt__', function (self, other) {
    return unpack_number(self) > unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__ge__', function (self, other) {
    return unpack_number(self) <= unpack_number(other) ? True : False;
}, ['other']);

py_float.$def('__pow__', function (self, other) {
    return pack_float(Math.pow(unpack_number(self), unpack_number(other)));
}, ['other']);
py_float.$def_alias('__pow__', '__ipow__');
py_float.$def_alias('__pow__', '__rpow__');

py_float.$def('__mul__', function (self, other) {
    return pack_float(unpack_number(self) * unpack_number(other));
}, ['other']);
py_float.$def_alias('__mul__', '__imul__');
py_float.$def_alias('__mul__', '__rmul__');

py_float.$def('__floordiv__', function (self, other) {
    return pack_int(Math.floor(unpack_number(self) / unpack_number(other)));
}, ['other']);
py_float.$def_alias('__floordiv__', '__ifloordiv__');
py_float.$def_alias('__floordiv__', '__rfloordiv__');

py_float.$def('__truediv__', function (self, other) {
    return pack_float(unpack_number(self) / unpack_number(other));
}, ['other']);
py_float.$def_alias('__truediv__', '__itruediv__');
py_float.$def_alias('__truediv__', '__rtruediv__');

py_float.$def('__mod__', function (self, other) {
    return pack_float(unpack_number(self) % unpack_number(other));
}, ['other']);
py_float.$def_alias('__mod__', '__imod__');
py_float.$def_alias('__mod__', '__rmod__');

py_float.$def('__add__', function (self, other) {
    return pack_float(unpack_number(self) + unpack_number(other));
}, ['other']);
py_float.$def_alias('__add__', '__iadd__');
py_float.$def_alias('__add__', '__radd__');

py_float.$def('__sub__', function (self, other) {
    return pack_float(unpack_number(self) - unpack_number(other));
}, ['other']);
py_float.$def_alias('__sub__', '__isub__');
py_float.$def_alias('__sub__', '__rsub__');
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_str.$def('__new__', function (cls, initializer, state, frame) {
    switch (state) {
        case 0:
            if (!issubclass(cls, py_str)) {
                raise(TypeError, 'class is not an subclass of str');
            }
            if (initializer instanceof PyStr) {
                if (initializer.cls == cls) {
                    return initializer;
                } else {
                    return pack_str(initializer.value, cls);
                }
            }
            if (initializer.call('__str__')) {
                return 1;
            }
        case 1:
            if (except(MethodNotFoundError)) {
                raise(TypeError, 'invalid type of str initializer');
            } else if (vm.return_value) {
                return vm.return_value;
            }
            break;
    }
}, ['initializer']);

py_str.$def('__str__', function (self) {
    return self;
});

py_str.$def('__repr__', function (self) {
    return self.repr();
});

py_str.$def('__len__', function (self)  {
    py_str.check(self);
    return pack_int(self.value.length);
});

py_str.$def('__add__', function (self, other) {
    return pack_str(unpack_str(self) + unpack_str(other));
}, ['other']);
py_str.$def_alias('__add__', '__iadd__');
py_str.$def_alias('__add__', '__radd__');

py_str.$def('__hash__', function (self) {
    return self;
});

py_str.$def('startswith', function (self, prefix) {
    return unpack_str(self).indexOf(unpack_str(prefix)) == 0 ? True : False;
}, ['prefix']);

py_str.$def('split', function (self, sep) {
    py_str.check(self);
    return new PyList(self.split(unpack_str(sep)).map(pack_str));
}, ['sep']);

function str(object) {
    return py_str.call_classmethod('__new__', [object]);
}

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_bytes.$def('decode', function (self, encoding) {
    py_bytes.check(self);
    return pack_str(self.decode(unpack_str(encoding)));
}, ['encoding'], {defaults: {'encoding': None}});
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_list.$def('__repr__', function (self, state, frame) {
    while (true) {
        switch (state) {
            case 0:
                py_list.check(self);
                frame.parts = new Array(self.size);
                frame.index = 0;
            case 1:
                if (frame.index < self.size) {
                    if (self.get(frame.index).call('__repr__')) {
                        return 2;
                    }
                } else {
                    state = 3;
                    break;
                }
            case 2:
                if (!vm.return_value) {
                    return;
                }
                frame.parts[frame.index] = vm.return_value.toString();
                frame.index++;
                state = 1;
                break;
            case 3:
                return pack_str('[' + frame.parts.join(', ') + ']');
        }
    }
});

py_list.$def('append', function (self, item) {
    py_list.check(self);
    self.append(item);
}, ['item']);


py_list.$def('__getitem__', function (self, index_or_slice) {
    py_list.check(self);
    // TODO: do conversion with __index__ and support slice
    return self.get(unpack_int(index_or_slice));
}, ['index_or_slice']);

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_function.$def('__get__', function (self, instance, owner) {
    return new PyMethod(instance, self);
}, ['instance', 'owner']);
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_generator.$def('__iter__', function (self) {
    return self;
});

py_generator.$def('__next__', function (self, state, frame) {
    switch (state) {
        case 0:
            py_generator.check(self);
            self.running = true;
            self.frame.back = frame;
            vm.frame = self.frame;
            vm.return_value = None;
            return 1;
        case 1:
            if (self.frame.why == CAUSES.YIELD) {
                return vm.return_value;
            } else if (self.frame.why == CAUSES.RETURN) {
                raise(StopIteration, new PyObject(StopIteration, {'args': pack_tuple([vm.return_value])}));
            }
    }
});
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function PyMethod(self, func) {
    PyObject.call(this, py_method);
    this.self = self;
    this.func = func;
}
PyMethod.prototype = new PyObject;

function new_method(func, instance) {
    return None;
}

py_method.$def('__str__', function (self) {
    return pack_str('<bound-method');
});

py_classmethod.$def('__init__', function (self, func) {
    self.setattr('__func__', func);
}, ['func']);

py_classmethod.$def('__get__', function (self, instance, owner) {
    return new_method(self.getattr('__func__'), owner);
}, ['instance', 'owner']);
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_property.$def('__get__', function (self, instance, owner, state, frame) {
    switch (state) {
        case 0:
            if (call(self.getattr('fget'), [instance])) {
                return 1;
            }
        case 1:
            return vm.return_value;
    }
}, ['instance', 'owner']);

py_property.$def('__set__', function (self, instance, value, state, frame) {
    switch (state) {
        case 0:
            if (call(self.getattr('fset'), [instance, value])) {
                return 1;
            }
        case 1:
            break;
    }
}, ['instance', 'value']);
/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

py_slice.$def('__repr__', function (self, state, frame) {
    switch (state) {
        case 0:
            py_slice.check(self);
            frame.result = 'slice(';
            if (self.start.call('__repr__')) {
                return 1;
            }
        case 1:
            if (!vm.return_value) {
                return;
            }
            frame.result += unpack_str(vm.return_value) + ', ';
            if (self.stop.call('__repr__')) {
                return 2;
            }
        case 2:
            if (!vm.return_value) {
                return;
            }
            frame.result += unpack_str(vm.return_value) + ', ';
            if (self.step.call('__repr__')) {
                return 3;
            }
        case 3:
            if (!vm.return_value) {
                return;
            }
            frame.result += unpack_str(vm.return_value) + ')';
            return pack_str(frame.result);
    }
});

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function PyModule(namespace) {
    PyObject.call(this, py_module, namespace);
}

PyModule.prototype = new PyObject;

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var NONE_STR = pack_str('None');

None.cls.$def('__new__', function (cls) {
    return None;
});

None.cls.$def('__repr__', function (self) {
    return NONE_STR;
});

None.cls.$def('__bool__', function (self) {
    return False;
});

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

BaseException.$def('__new__', function (cls, args) {
    BaseException.check_subclass(cls);
    return new PyObject(cls, {'args': pack_tuple(args)})
}, ['*args']);

Exception.$def('__init__', function (self, args) {
    self.setattr('args', pack_tuple(args));
}, ['*args']);

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var build_class = $def(function (func, name, bases, metaclass, keywords, state, frame) {
    var possible_meta_classes, index, good, bases;
    if (!(func.cls == py_function)) {
        raise(TypeError, 'invalid type of \'func\' argument');
    }
    if (!(name instanceof PyStr)) {
        raise(TypeError, 'invalid type of \'name\' argument');
    }
    if (!(bases instanceof Array)) {
        raise(TypeError, 'invalid type of \'bases\' argument');
    }
    switch (state) {
        case 0:
            if (metaclass === None && bases.length == 0) {
                frame.metaclass = py_type;
            } else if (!(metaclass instanceof PyType)) {
                frame.metaclass = metaclass;
            } else {
                possible_meta_classes = [];
                if (metaclass !== None) {
                    possible_meta_classes.push(metaclass);
                }
                for (index = 0; index < bases.length; index++) {
                    if (bases.value[index] instanceof PyType) {
                        possible_meta_classes.push(bases[index].cls)
                    } else {
                        raise(TypeError, 'invalid type of base');
                    }
                }
                for (index = 0; index < possible_meta_classes.length; index++) {
                    good = true;
                    possible_meta_classes.forEach(function (meta_class) {
                        if (!issubclass(possible_meta_classes[index], meta_class)) {
                            good = false;
                        }
                    });
                    if (good) {
                        break;
                    }
                }
                if (good) {
                    frame.metaclass = possible_meta_classes[index];
                } else {
                    raise(TypeError, 'unable to determine most derived metaclass');
                }
            }
            if (frame.metaclass.call_classmethod('__prepare__', [pack_tuple(bases)], keywords)) {
                return 1;
            }
        case 1:
            if (!vm.return_value) {
                return null;
            }
            frame.dict = vm.return_value;
            if (call(func, undefined, undefined, undefined, undefined, undefined, frame.dict)) {
                return 2;
            }
        case 2:
            if (!vm.return_value) {
                return null;
            }
            if (bases.length == 0) {
                bases = [py_object];
            } else {
                bases = bases.value;
            }
            if (frame.metaclass.cls.call('__call__', [name, pack_tuple(bases), frame.dict], keywords)) {
                return 3;
            }
        case 3:
            if (!vm.return_value) {
                return null;
            }
            return vm.return_value;
    }
}, ['func', 'name', '*bases', 'metaclass', '**keywords'], {
    name: '__build_class__',
    defaults: {'metaclass': None}
});

var builtins = {
    object: py_object,
    type: py_type,
    dict: py_dict,
    int: py_int,
    float: py_float,
    str: py_str,
    bytes: py_bytes,
    tuple: py_tuple,

    None: None,
    NotImplemented: NotImplemented,
    Ellipsis: Ellipsis,
    False: False,
    True: True,

    BaseException: BaseException,
    Exception: Exception,
    ValueError: ValueError,
    ArithmeticError: ArithmeticError,
    LookupError: LookupError,
    RuntimeError: RuntimeError,
    BufferError: BufferError,
    AssertionError: AssertionError,
    AttributeError: AttributeError,
    EOFError: EOFError,
    FloatingPointError: FloatingPointError,
    GeneratorExit: GeneratorExit,
    ImportError: ImportError,
    IndexError: IndexError,
    KeyError: KeyError,
    KeyboardInterrupt: KeyboardInterrupt,
    MemoryError: MemoryError,
    NameError: NameError,
    NotImplementedError: NotImplementedError,
    OSError: OSError,
    OverflowError: OverflowError,
    RecursionError: RecursionError,
    ReferenceError: ReferenceError,
    StopIteration: StopIteration,
    SyntaxError: SyntaxError,
    IndentationError: IndentationError,
    TabError: TabError,
    SystemError: SystemError,
    SystemExit: SystemExit,
    TypeError: TypeError,
    UnboundLocalError: UnboundLocalError,
    UnicodeError: UnicodeError,
    UnicodeEncodeError: UnicodeEncodeError,
    UnicodeDecodeError: UnicodeDecodeError,
    UnicodeTranslateError: UnicodeTranslateError,
    ZeroDivisionError: ZeroDivisionError,
    EnvironmentError: EnvironmentError,
    BlockingIOError: BlockingIOError,
    ChildProcessError: ChildProcessError,
    BrokenPipeError: BrokenPipeError,
    ConnectionError: ConnectionError,
    ConnectionAbortedError: ConnectionAbortedError,
    ConnectionRefusedError: ConnectionRefusedError,
    ConnectionResetError: ConnectionResetError,
    FileExistsError: FileExistsError,
    FileNotFoundError: FileNotFoundError,
    InterruptedError: InterruptedError,
    IsADirectoryError: IsADirectoryError,
    NotADirectoryError: NotADirectoryError,
    PermissionError: PermissionError,
    ProcessLookupError: ProcessLookupError,
    TimeoutError: TimeoutError,

    __build_class__: build_class
};


var module_builtins = module('builtins', function ($, module) {
    module.dict = builtins;
});

module('js', function ($, module) {
    return {
        'JSError': JSError
    }
});

module_builtins.$def('__import__', function (name, globals, locals, fromlist, level, state, frame) {
    name = unpack_str(name);
    switch (state) {
        case 0:
            if (level.gt(ZERO)) {
                raise(ImportError, 'relative imports are not supported')
            }
            if (!(name in modules)) {
                raise(ImportError, 'unable to import module ' + name);
            }
            if (modules[name] instanceof NativeModule) {
                return new PyModule(modules[name].dict);
            }
            if (modules[name].wrapper) {
                return modules[name].wrapper;
            }
            if (call(modules[name])) {
                return 1;
            }
        case 1:
            var module = new PyModule(modules[name].dict);
            modules[name].wrapper = module;
            return module;
    }
}, ['name', 'globals', 'locals', 'fromlist', 'level'], {defaults: {
    'globals': None, 'locals': None, 'fromlist': EMPTY_TUPLE, 'level': ZERO
}});

module_builtins.$def('print', function (objects, sep, end, file, flush, state, frame) {
    var object;
    while (true) {
        switch (state) {
            case 0:
                frame.strings = [];
                frame.index = 0;
                if (objects.length) {
                    object = objects[0];
                    if (object.cls === py_str) {
                        vm.return_value = object;
                    } else if (object.call('__str__')) {
                        return 1;
                    }
                    state = 1;
                } else {
                    state = 2;
                    break;
                }
            case 1:
                if (!vm.return_value) {
                    return null;
                }
                frame.strings.push(unpack_str(vm.return_value));
                frame.index++;
                if (frame.index < objects.length) {
                    object = objects[frame.index];
                    if (object.cls == py_str) {
                        vm.return_value = object;
                    } else if (object.call('__str__')) {
                        return 1;
                    }
                    break;
                }
            case 2:
                if (sep.cls == py_str) {
                    vm.return_value = sep;
                } else if (sep.call('__str__')) {
                    return 3;
                }
            case 3:
                if (!vm.return_value) {
                    return null;
                }
                console.log(frame.strings.join(unpack_str(vm.return_value)));
                return null;
        }
    }
}, ['*objects', 'sep', 'end', 'file', 'flush'], {
    defaults: {sep: pack_str(' '), end: pack_str('\n'), file: None, flush: False}
});

module_builtins.$def('len', function (object, state, frame) {
    switch (state) {
        case 0:
            if (object.call('__len__')) {
                return 1;
            }
        case 1:
            return vm.return_value;
    }

}, ['object']);

var getattr = module_builtins.$def('getattr', function (object, name, state, frame) {
    switch (state) {
        case 0:
            if (object.call('__getattribute__', [name])) {
                return 1;
            }
        case 1:
            if (vm.return_value) {
                return vm.return_value;
            }
            if (except(AttributeError) || except(MethodNotFoundError)) {
                if (object.call('__getattr__', [name])) {
                    return 2;
                }
            } else {
                return;
            }
        case 2:
            if (except(MethodNotFoundError)) {
                raise(TypeError, 'object does not support attribute access');
            }
            return vm.return_value;
    }
}, ['object', 'name']);



$.builtins = builtins;

$.None = None;
$.NotImplemented = NotImplemented;
$.Ellipsis = Ellipsis;
$.False = False;
$.True = True;


/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

var unpack_sequence = $def(function (sequence, limit, state, frame) {
    while (true) {
        switch (state) {
            case 0:
                frame.result = [];
                frame.limit = unpack_int(limit, Infinity);
                if (sequence.cls.lookup('__next__')) {
                    frame.iterator = sequence;
                    state = 2;
                    break;
                }
                if (sequence.call('__iter__')) {
                    return 1;
                }
            case 1:
                if (!vm.return_value) {
                    if (except(MethodNotFoundError)) {
                        raise(TypeError, 'object does not support the iterable protocol');
                    }
                    return;
                }
                frame.iterator = vm.return_value;
            case 2:
                if (frame.iterator.call('__next__')) {
                    return 3;
                }
            case 3:
                if (!vm.return_value) {
                    if (except(StopIteration)) {
                        state = 4;
                        break;
                    } else if (except(MethodNotFoundError)) {
                        raise(TypeError, 'object does not support the iterable protocol');
                    }
                    return;
                }
                frame.result.push(vm.return_value);
                if (frame.result.length > frame.limit) {
                    raise(ValueError, 'too many values to unpack (expected ' + frame.limit + ')');
                }
                state = 2;
                break;
            case 4:
                return pack_tuple(frame.result);
        }
    }
}, ['sequence', 'limit'], {name: 'unpack_sequence', defaults: {'limit': None}});



    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

function disassemble(code) {
    var instruction, opcode, high, low, argument, index;

    var instructions = [];
    var table = {};
    var position = 0;

    while (position < code.bytecode.length) {
        instruction = {start: position};
        table[position] = instruction;

        opcode = code.bytecode.charCodeAt(position++);
        if (opcode >= OPCODES_ARGUMENT) {
            low = code.bytecode.charCodeAt(position++);
            high = code.bytecode.charCodeAt(position++);
            argument = high << 8 | low;
        }
        if (opcode === OPCODES.EXTENDED_ARG) {
            opcode = code.bytecode.charCodeAt(position++);
            low = code.bytecode.charCodeAt(position++);
            high = code.bytecode.charCodeAt(position++);
            argument = (argument << 16) | (high << 8) | low;
        }
        instruction.position = instructions.length;
        instruction.end = position;
        instruction.opcode = opcode;
        instruction.argument = argument;
        instructions.push(instruction);
    }

    for (index = 0; index < instructions.length; index++) {
        instruction = instructions[index];
        switch (instruction.opcode) {
            case OPCODES.JUMP_FORWARD:
                instruction.opcode = OPCODES.JUMP_ABSOLUTE;
            case OPCODES.FOR_ITER:
            case OPCODES.SETUP_EXCEPT:
            case OPCODES.SETUP_FINALLY:
            case OPCODES.SETUP_LOOP:
            case OPCODES.SETUP_WITH:
                instruction.target = table[instruction.end + instruction.argument].position;
                break;

            case OPCODES.JUMP_IF_FALSE_OR_POP:
            case OPCODES.JUMP_IF_TRUE_OR_POP:
            case OPCODES.POP_JUMP_IF_FALSE:
            case OPCODES.POP_JUMP_IF_TRUE:
            case OPCODES.JUMP_ABSOLUTE:
                instruction.target = table[instruction.argument].position;
                break;
        }
    }

    return instructions;
}


$.disassemble = disassemble;


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var Frame = Class.extend({
    constructor: function (code, options) {
        this.code = code;

        options = options || {};

        this.back = options.back || null;

        this.locals = {};
        this.globals = options.globals || (this.back ? this.back.globals : {});
        this.builtins = options.builtins || (this.back ? this.back.builtins : builtins);

        this.state = options.state || 0;

        
            this.thread = this.back ? this.back.thread : threading.thread;
        

        
            this.debug_line = -1;
            this.debug_break = false;

            this.debug_step_over = false;
            this.debug_step_into = false;
            this.debug_step_out = false;

            this.debug_future = null;

            this.debug_internal = options.debug_internal || false;
        
    }
});


var PythonFrame = Frame.extend({
    constructor: function (code, options) {
        var index;

        options = options || {};

        Frame.call(this, code, options);

        this.locals = options.locals || {};

        this.code.parse_args(options.args, options.kwargs, options.defaults, this.locals);

        this.namespace = options.namespace || null;

        this.stack = new Array(this.code.stacksize);
        this.level = 0;

        this.blocks = options.blocks || [];
        if (!this.blocks.length) {
            this.push_block(BLOCK_TYPES.BASE, 0);
        }

        this.position = options.position || 0;

        this.closure = options.closure || [];

        this.cells = {};
        for (index = 0; index < this.code.cellvars.length; index++) {
            this.cells[this.code.cellvars[index]] = new PyCell();
        }
        for (index = 0; index < this.code.freevars.length; index++) {
            this.cells[this.code.freevars[index]] = this.closure[index];
        }

        this.why = null;
    },

    top_block: function () {
        return this.blocks[this.blocks.length - 1];
    },

    push_block: function (type, target) {
        this.blocks.push({
            type: type,
            position: this.position,
            target: target,
            active: false,
            level: this.level
        });
    },

    pop_block: function () {
        return this.blocks.pop();
    },

    pop: function () {
        return this.stack[--this.level];
    },

    popn: function (number) {
        this.level -= number;
        return this.stack.slice(this.level, this.level + number);
    },

    top0: function () {
        return this.stack[this.level - 1];
    },

    top1: function () {
        return this.stack[this.level - 2];
    },

    topn: function (number) {
        return this.stack.slice(this.level - number, this.level);
    },

    peek: function (number) {
        return this.stack[this.level - number];
    },

    push: function (item) {
        assert(item instanceof PyObject, 'tried to push non python object on stack');
        this.stack[this.level++] = item;
        if (this.level > this.code.stacksize) {
            error('stack overflow');
        }
    },

    set_state: function (state) {
        this.position--;
        this.state = state;
    },

    print_block_stack: function () {
        for (var index = 0; index < this.blocks.length; index++) {
            console.log('Block ' + index + ':', this.blocks[index]);
        }
    },

    unwind: function (cause) {
        if (cause != undefined) {
            this.why = cause;
        }
        while (this.blocks.length > 0) {
            var block = this.blocks[this.blocks.length - 1];
            if (block.active) {
                this.blocks.pop();
                continue;
            }
            if (block.type == BLOCK_TYPES.FINALLY) {
                this.position = block.target;
                block.active = true;
                return;
            }
            switch (this.why) {
                case CAUSES.BREAK:
                    if (block.type == BLOCK_TYPES.LOOP) {
                        this.position = block.target;
                        this.blocks.pop();
                        return;
                    } else {
                        this.blocks.pop();
                    }
                    break;
                case CAUSES.CONTINUE:
                    if (block.type == BLOCK_TYPES.LOOP) {
                        this.position = block.position;
                        return;
                    } else {
                        this.blocks.pop();
                    }
                    break;
                case CAUSES.EXCEPTION:
                    if (block.type == BLOCK_TYPES.EXCEPT) {
                        this.position = block.target;
                        block.active = true;
                        return;
                    } else if (block.type == BLOCK_TYPES.BASE) {
                        vm.frame = this.back;
                        
                            if (!vm.frame) {
                                threading.finished();
                            }
                        
                        return;
                    } else {
                        this.blocks.pop()
                    }
                    break;
                case CAUSES.RETURN:
                    if (block.type == BLOCK_TYPES.BASE) {
                        vm.frame = this.back;
                        
                            debugging.trace_return(this);
                        
                        
                            if (!vm.frame) {
                                threading.finished();
                            }
                        
                        return;
                    } else {
                        this.blocks.pop();
                    }
                    break;
                default:
                    error('unknown unwind cause ' + this.why);
            }
        }
    },

    raise: function () {
        this.push(vm.last_exception.exc_tb);
        this.push(vm.last_exception.exc_value);
        this.push(vm.last_exception.exc_type);
        this.unwind(CAUSES.EXCEPTION);
    },

    get_line_number: function () {
        return this.code.get_line_number(this.position - 1);
    }
});


var NativeFrame = Frame.extend({
    constructor: function (code, options) {
        options = options || {};
        Frame.call(this, code, options);
        this.args = this.code.parse_args(options.args, options.kwargs, options.defaults);
    },

    get_line_number: function () {
        return this.state;
    }
});




$.Frame = Frame;
$.PythonFrame = PythonFrame;
$.NativeFrame = NativeFrame;

    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */







PythonFrame.prototype.execute = function() {
    var slot, right, left, name, key, value, block, exc_type, exc_value, exc_tb, temp;
    var low, mid, high, args, kwargs, index, code, defaults, globals, func, instruction;

    while (vm.frame === this) {
        if (!vm.return_value && this.why != CAUSES.EXCEPTION && this.state == 0) {
            this.raise();
        }

        
            if (threading.step()) {
                return;
            }
        

        
            var line = this.get_line_number();
            if (line != this.debug_line) {
                debugging.trace_line(this, line);
                this.debug_line = line;
            }

            if (debugging.step()) {
                return;
            }
        

        instruction = this.code.instructions[this.position++];

        

        switch (instruction.opcode) {
            case OPCODES.NOP:
                break;

            case OPCODES.POP_TOP:
                this.pop();
                break;

            case OPCODES.ROT_TWO:
                temp = this.popn(2);
                this.push(temp[1]);
                this.push(temp[0]);
                break;

            case OPCODES.ROT_THREE:
                temp = this.popn(3);
                this.push(temp[2]);
                this.push(temp[1]);
                this.push(temp[0]);
                break;

            case OPCODES.DUP_TOP:
                this.push(this.top0());
                break;

            case OPCODES.DUP_TOP_TWO:
                this.push(this.top1());
                this.push(this.top1());
                break;

            case OPCODES.GET_YIELD_FROM_ITER:
                if (this.state === 0 && isiterable(this.top0())) {
                    break;
                }
            case OPCODES.UNARY_POSITIVE:
            case OPCODES.UNARY_NEGATIVE:
            case OPCODES.UNARY_NOT:
            case OPCODES.UNARY_INVERT:
            case OPCODES.GET_ITER:
                switch (this.state) {
                    case 0:
                        if (this.pop().call(OPCODES_EXTRA[instruction.opcode])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                            raise(TypeError, 'unsupported operand type');
                        } else if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.BINARY_POWER:
            case OPCODES.BINARY_MULTIPLY:
            case OPCODES.BINARY_MATRIX_MULTIPLY:
            case OPCODES.BINARY_FLOOR_DIVIDE:
            case OPCODES.BINARY_TRUE_DIVIDE:
            case OPCODES.BINARY_MODULO:
            case OPCODES.BINARY_ADD:
            case OPCODES.BINARY_SUBTRACT:
            case OPCODES.BINARY_SUBSCR:
            case OPCODES.BINARY_LSHIFT:
            case OPCODES.BINARY_RSHIFT:
            case OPCODES.BINARY_AND:
            case OPCODES.BINARY_XOR:
            case OPCODES.BINARY_OR:
                switch (this.state) {
                    case 0:
                        slot = OPCODES_EXTRA[instruction.opcode];
                        right = this.top0();
                        left = this.top1();
                        if (left.call('__' + slot + '__', [right])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                            slot = OPCODES_EXTRA[instruction.opcode];
                            right = this.pop();
                            left = this.pop();
                            if (right.call('__r' + slot + '__', [left])) {
                                return 2;
                            }
                        } else {
                            this.popn(2);
                            if (vm.return_value) {
                                this.push(vm.return_value);
                            }
                            break;
                        }
                    case 2:
                        if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                            raise(TypeError, 'unsupported operand type');
                        } else if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.INPLACE_POWER:
            case OPCODES.INPLACE_MULTIPLY:
            case OPCODES.INPLACE_MATRIX_MULTIPLY:
            case OPCODES.INPLACE_FLOOR_DIVIDE:
            case OPCODES.INPLACE_TRUE_DIVIDE:
            case OPCODES.INPLACE_MODULO:
            case OPCODES.INPLACE_ADD:
            case OPCODES.INPLACE_SUBTRACT:
            case OPCODES.INPLACE_LSHIFT:
            case OPCODES.INPLACE_RSHIFT:
            case OPCODES.INPLACE_AND:
            case OPCODES.INPLACE_XOR:
            case OPCODES.INPLACE_OR:
            case OPCODES.DELETE_SUBSCR:
                switch (this.state) {
                    case 0:
                        right = this.pop();
                        left = this.pop();
                        if (left.call(OPCODES_EXTRA[instruction.opcode], [right])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                            raise(TypeError, 'unsupported operand type');
                        } else if (vm.return_value && instruction.opcode != OPCODES.DELETE_SUBSCR) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.STORE_SUBSCR:
                switch (this.state) {
                    case 0:
                        name = this.pop();
                        left = this.pop();
                        right = this.pop();
                        if (left.call('__setitem__', [name, right])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                            raise(TypeError, 'unsupported operand type');
                        }
                        break;
                }
                break;

            case OPCODES.PRINT_EXPR:
                console.log(this.pop());
                break;

            case OPCODES.BREAK_LOOP:
                this.unwind(CAUSES.BREAK);
                break;

            case OPCODES.CONTINUE_LOOP:
                this.unwind(CAUSES.CONTINUE);
                break;

            case OPCODES.SET_ADD:
                switch (this.state) {
                    case 0:
                        value = this.pop();
                        if (this.peek(instruction.argument).call('add', [value])) {
                            return 1;
                        }
                    case 1:
                        break;
                }
                break;

            case OPCODES.LIST_APPEND:
                switch (this.state) {
                    case 0:
                        value = this.pop();
                        if (this.peek(instruction.argument).call('append', [value])) {
                            return 1;
                        }
                    case 1:
                        break;
                }
                break;

            case OPCODES.MAP_ADD:
                switch (this.state) {
                    case 0:
                        key = this.pop();
                        value = this.pop();
                        if (this.peek(instruction.argument).call('__setitem__', [key, value])) {
                            return 1;
                        }
                    case 1:
                        break;
                }
                break;

            case OPCODES.RETURN_VALUE:
                vm.return_value = this.pop();
                this.unwind(CAUSES.RETURN);
                break;

            case OPCODES.YIELD_VALUE:
                switch (this.state) {
                    case 0:
                        vm.return_value = this.pop();
                        vm.frame = this.back;
                        this.why = CAUSES.YIELD;
                        return 1;
                    case 1:
                        this.push(vm.return_value);
                        this.why = CAUSES.RUN;
                        break;
                }
                break;

            case OPCODES.YIELD_FROM:
                error('opcode not implemented');
                break;

            case OPCODES.POP_BLOCK:
                this.blocks.pop();
                break;

            case OPCODES.POP_EXCEPT:
                block = this.blocks.pop();
                assert(block.type === BLOCK_TYPES.EXCEPT, 'invalid type of top block');
                while (this.level > block.level + 3) {
                    this.pop();
                }
                if (this.level == block.level + 3) {
                    exc_type = this.pop();
                    exc_value = this.pop();
                    exc_tb = this.pop();
                    raise(exc_type, exc_value, exc_tb);
                } else {
                    vm.return_value = None;
                    this.why = CAUSES.RUN;
                }
                break;

            case OPCODES.END_FINALLY:
                this.unwind();
                break;

            case OPCODES.LOAD_BUILD_CLASS:
                this.push(build_class);
                break;

            case OPCODES.SETUP_WITH:
                switch (this.state) {
                    case 0:
                        if (call(getattr, [this.top0(), pack_str('__exit__')])) {
                            return 1;
                        }
                    case 1:
                        temp = this.pop();
                        if (!vm.return_value) {
                            break;
                        }
                        this.push(vm.return_value);
                        if (temp.call('__enter__')) {
                            return 2;
                        }
                    case 2:
                        if (!vm.return_value) {
                            this.pop();
                            break;
                        }
                        this.push(vm.return_value);
                        this.push_block(BLOCK_TYPES.FINALLY, instruction.target);
                }
                break;

            case OPCODES.WITH_CLEANUP_START:
                switch (this.state) {
                    case 0:
                        switch (this.why) {
                            case CAUSES.EXCEPTION:
                                args = [this.pop(), this.pop(), this.pop()];
                                break;
                            case CAUSES.RETURN:
                            case CAUSES.CONTINUE:
                            case CAUSES.BREAK:
                                args = [None, None, None];
                                break;
                            default:
                                args = [None, None, None];
                                this.pop();
                        }
                        this.return_value = vm.return_value;
                        vm.return_value = None;
                        if (call(this.pop(), args)) {
                            return 1;
                        }
                    case 1:
                        if (!vm.return_value) {
                            break;
                        }
                        this.push(vm.return_value);
                }
                break;

            case OPCODES.WITH_CLEANUP_FINISH:
                if (this.pop().bool()) {
                    if (this.why == CAUSES.EXCEPTION) {
                        this.why = CAUSES.RUN;
                    }
                } else {
                    if (this.why == CAUSES.EXCEPTION) {
                        this.push(vm.last_exception.exc_tb);
                        this.push(vm.last_exception.exc_value);
                        this.push(vm.last_exception.exc_type);
                    }
                }
                if (this.why == CAUSES.RETURN) {
                    vm.return_value = this.return_value;
                }
                break;

            case OPCODES.STORE_FAST:
            case OPCODES.DELETE_FAST:
            case OPCODES.STORE_NAME:
            case OPCODES.DELETE_NAME:
                if (instruction.opcode === OPCODES.STORE_FAST) {
                    name = this.code.varnames[instruction.argument];
                } else if (instruction.opcode === OPCODES.DELETE_FAST) {
                    name = this.code.varnames[instruction.argument];
                } else {
                    name = this.code.names[instruction.argument];
                }
                if (this.namespace) {
                    switch (this.state) {
                        case 0:
                            if (instruction.opcode === OPCODES.STORE_NAME) {
                                slot = '__setitem__';
                                args = [pack_str(name), this.pop()];
                            } else {
                                slot = '__delitem__';
                                args = [pack_str(name)];
                            }
                            if (this.namespace.call(slot, args)) {
                                return 1;
                            }
                        case 1:
                            if (except(MethodNotFoundError)) {
                                raise(TypeError, 'invalid namespace');
                            }
                            break;
                    }
                } else {
                    if (instruction.opcode === OPCODES.STORE_NAME || instruction.opcode === OPCODES.STORE_FAST) {
                        this.locals[name] = this.pop();
                    } else if (name in this.locals) {
                        delete this.locals[name];
                    } else {
                        raise(NameError, 'name \'' + name + '\' is not defined');
                    }
                }
                break;

            case OPCODES.LOAD_FAST:
            case OPCODES.LOAD_NAME:
                if (instruction.opcode === OPCODES.LOAD_FAST) {
                    name = this.code.varnames[instruction.argument];
                } else {
                    name = this.code.names[instruction.argument];
                }
                if (this.namespace) {
                    switch (this.state) {
                        case 0:
                            if (this.namespace.call('__getitem__', [pack_str(name)])) {
                                return 1;
                            }
                        case 1:
                            if (vm.return_value) {
                                this.push(vm.return_value);
                            } else if (except(MethodNotFoundError) || except(KeyError)) {
                                if (name in this.globals) {
                                    this.push(this.globals[name]);
                                } else if (name in this.builtins) {
                                    this.push(this.builtins[name]);
                                } else {
                                    raise(NameError, 'name \'' + name + '\' is not defined');
                                }
                            }
                            break;
                    }
                } else {
                    if (name in this.locals) {
                        this.push(this.locals[name]);
                    } else if (name in this.globals) {
                        this.push(this.globals[name]);
                    } else if (name in this.builtins) {
                        this.push(this.builtins[name]);
                    } else {
                        raise(NameError, 'name \'' + name + '\' is not defined');
                    }
                }
                break;

            case OPCODES.STORE_GLOBAL:
                name = this.code.names[instruction.argument];
                this.globals[name] = this.pop();
                break;

            case OPCODES.DELETE_GLOBAL:
                name = this.code.names[instruction.argument];
                if (name in this.globals) {
                    delete this.globals[name];
                } else {
                    raise(NameError, 'name \'' + name + '\' is not defined');
                }
                break;

            case OPCODES.LOAD_GLOBAL:
                name = this.code.names[instruction.argument];
                if (name in this.globals) {
                    this.push(this.globals[name]);
                } else if (name in this.builtins) {
                    this.push(this.builtins[name]);
                } else {
                    raise(NameError, 'name \'' + name + '\' is not defined');
                }
                break;

            case OPCODES.STORE_ATTR:
            case OPCODES.DELETE_ATTR:
                name = this.code.names[instruction.argument];
                switch (this.state) {
                    case 0:
                        temp = this.pop();
                        if (instruction.opcode === OPCODES.STORE_ATTR) {
                            slot = '__setattr__';
                            args = [pack_str(name), this.pop()];
                        } else {
                            slot = '__delattr__';
                            args = [pack_str(name)];
                        }
                        if (temp.call(slot, args)) {
                            return 1;
                        }
                    case 1:
                        if (except(MethodNotFoundError)) {
                            if (instruction.opcode === OPCODES.STORE_ATTR) {
                                raise(TypeError, 'object does not support attribute assignment');
                            } else {
                                raise(TypeError, 'object does not support attribute deletion');
                            }
                        }
                        break;
                }
                break;

            case OPCODES.LOAD_ATTR:
                name = this.code.names[instruction.argument];
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__getattribute__', [pack_str(name)])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value) {
                            this.pop();
                            this.push(vm.return_value);
                            break;
                        }
                        if (except(AttributeError) || except(MethodNotFoundError)) {
                            if (this.pop().call('__getattr__', [pack_str(name)])) {
                                return 2;
                            }
                        } else {
                            this.pop();
                            break;
                        }
                    case 2:
                        if (except(MethodNotFoundError)) {
                            raise(TypeError, 'object does not support attribute access');
                        } else if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.UNPACK_SEQUENCE:
                switch (this.state) {
                    case 0:
                        if (call(unpack_sequence, [this.pop(), pack_int(instruction.argument)])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value) {
                            if (vm.return_value.len() != instruction.argument) {
                                raise(TypeError, 'not enough values to unpack (expected ' + instruction.argument + ', got ' + vm.return_value.len() + ')');
                            } else {
                                for (index = vm.return_value.len() - 1; index >= 0; index--) {
                                    this.push(vm.return_value.get(index));
                                }
                            }
                        }
                }
                break;

            case OPCODES.UNPACK_EX:
                switch (this.state) {
                    case 0:
                        if (call(unpack_sequence, [this.pop()])) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value) {
                            low = instruction.argument & 0xFF;
                            high = instruction.argument >> 8;
                            if (vm.return_value.len() < low + high) {
                                raise(TypeError, 'not enough values to unpack (expected at least' + (low + high) + ', got ' + vm.return_value.len() + ')');
                            }
                            temp = vm.return_value.len() - high;
                            for (index = vm.return_value.len() - 1; index >= temp; index--) {
                                this.push(vm.return_value.get(index));
                            }
                            this.push(new PyList(vm.return_value.array.slice(low, temp)));
                            for (index = low - 1; index >= 0; index--) {
                                this.push(vm.return_value.get(index));
                            }
                        }
                }
                break;

            case OPCODES.BUILD_TUPLE:
                this.push(pack_tuple(this.popn(instruction.argument)));
                break;

            case OPCODES.BUILD_LIST:
                this.push(new PyList(this.popn(instruction.argument)));
                break;

            case OPCODES.BUILD_SET:
                error('opcode not implemented');
                break;

            case OPCODES.BUILD_MAP:
                this.push(new PyDict());
                break;

            case OPCODES.IMPORT_NAME:
                switch (this.state) {
                    case 0:
                        name = this.code.names[instruction.argument];
                        if (!('__import__' in this.builtins)) {
                            raise(ImportError, '__import__ not found');
                            break;
                        } else if (call(this.builtins['__import__'], [pack_str(name), new PyDict(this.globals), new PyDict(this.locals)].concat(this.popn(2).reverse()))) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                }
                break;

            case OPCODES.IMPORT_STAR:
                temp = this.pop();
                for (name in temp.dict) {
                    if (name.indexOf('_') != 0 && temp.dict.hasOwnProperty(name)) {
                        this.locals[name] = temp.dict[name];
                    }
                }
                break;

            case OPCODES.IMPORT_FROM:
                name = this.code.names[instruction.argument];
                if (name in this.top0().dict) {
                    this.push(this.top0().dict[name])
                } else {
                    raise(ImportError, 'cannot import name \'' + name + '\'');
                }
                break;

            case OPCODES.JUMP_FORWARD:
                this.position = instruction.target;
                break;

            case OPCODES.JUMP_ABSOLUTE:
                this.position = instruction.target;
                break;

            case OPCODES.COMPARE_OP:
                switch (instruction.argument) {
                    case COMPARE_OPS.EXC:
                        exc_type = this.pop();
                        this.push(issubclass(this.pop(), exc_type) ? True : False);
                        break;
                    case COMPARE_OPS.LT:
                    case COMPARE_OPS.LE:
                    case COMPARE_OPS.GT:
                    case COMPARE_OPS.GE:
                    case COMPARE_OPS.EQ:
                    case COMPARE_OPS.NE:
                        switch (this.state) {
                            case 0:
                                slot = COMPARE_SLOTS[instruction.argument];
                                right = this.pop();
                                left = this.pop();
                                if (left.call(slot, [right])) {
                                    return 1;
                                }
                            case 1:
                                if (vm.return_value === NotImplemented || except(MethodNotFoundError)) {
                                    raise(TypeError, 'unsupported boolean operator');
                                } else if (vm.return_value) {
                                    this.push(vm.return_value);
                                }
                                break;
                        }
                        break;

                    case COMPARE_OPS.IS:
                    case COMPARE_OPS.NIS:
                        right = this.pop();
                        left = this.pop();
                        if (right.is(left)) {
                            this.push(instruction.argument == COMPARE_OPS.IS ? True : False)
                        } else {
                            this.push(instruction.argument == COMPARE_OPS.NIS ? False : True)
                        }
                        break;

                    default:
                        error('unsupported comparison operator');
                }
                break;

            case OPCODES.POP_JUMP_IF_TRUE:
                if (this.top0().cls === py_int) {
                    if (this.pop().bool()) {
                        this.position = instruction.target;
                    }
                    break;
                }
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__bool__')) {
                            return 1;
                        }
                    case 1:
                        if (except(MethodNotFoundError)) {
                            if (this.pop().call('__len__')) {
                                return 2;
                            }
                        }
                    case 2:
                        if (!vm.return_value) {
                            break;
                        }
                        if (except(MethodNotFoundError)) {
                            this.position = instruction.target;
                            break;
                        } else if (vm.return_value) {
                            if (vm.return_value instanceof PyInt) {
                                if (vm.return_value.bool()) {
                                    this.position = instruction.target;
                                    break;
                                }
                            } else {
                                raise(TypeError, 'invalid result type of boolean conversion');
                            }
                        }
                        break;
                }
                break;

            case OPCODES.POP_JUMP_IF_FALSE:
                if (this.top0().cls === py_bool || this.top0().cls === py_int) {
                    if (!this.pop().bool()) {
                        this.position = instruction.target;
                    }
                    break;
                }
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__bool__')) {
                            return 1;
                        }
                    case 1:
                        if (except(MethodNotFoundError)) {
                            if (this.top0().call('__len__')) {
                                return 2;
                            }
                        } else if (!vm.return_value) {
                            this.pop();
                            break;
                        }
                    case 2:
                        this.pop();
                        if (!except(MethodNotFoundError)) {
                            if (vm.return_value instanceof PyInt) {
                                if (!vm.return_value.bool()) {
                                    this.position = instruction.target;
                                    break;
                                }
                            } else if (vm.return_value) {
                                raise(TypeError, 'invalid result type of boolean conversion');
                            }
                        }
                        break;
                }
                break;

            case OPCODES.JUMP_IF_TRUE_OR_POP:
                if (this.top0().cls === py_int) {
                    if (this.top0().bool()) {
                        this.position = instruction.target;
                    } else {
                        this.pop();
                    }
                    break;
                }
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__bool__')) {
                            return 1;
                        }
                    case 1:
                        if (except(MethodNotFoundError)) {
                            if (this.top0().call('__len__')) {
                                return 2;
                            }
                        } else if (!vm.return_value) {
                            this.pop();
                            break;
                        }
                    case 2:
                        if (except(MethodNotFoundError)) {
                            this.position = instruction.target;
                        } else if (vm.return_value) {
                            if (vm.return_value instanceof PyInt) {
                                if (vm.return_value.bool()) {
                                    this.position = instruction.target;
                                } else {
                                    this.pop();
                                }
                            } else {
                                this.pop();
                                raise(TypeError, 'invalid result type of boolean conversion');
                            }
                        } else {
                            this.pop();
                        }
                        break;
                }
                break;

            case OPCODES.JUMP_IF_FALSE_OR_POP:
                if (this.top0().cls === py_int) {
                    if (!this.top0().bool()) {
                        this.position = instruction.target;
                    } else {
                        this.pop();
                    }
                    break;
                }
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__bool__')) {
                            return 1;
                        }
                    case 1:
                        if (except(MethodNotFoundError)) {
                            if (this.top0().call('__len__')) {
                                return 2;
                            }
                        } else if (!vm.return_value) {
                            this.pop();
                            this.raise();
                            break;
                        }
                    case 2:
                        if (!except(MethodNotFoundError)) {
                            if (vm.return_value instanceof PyInt) {
                                if (!vm.return_value.bool()) {
                                    this.position = instruction.target;
                                } else {
                                    this.pop();
                                }
                            } else if (vm.return_value) {
                                this.pop();
                                raise(TypeError, 'invalid result type of boolean conversion');
                                this.raise();
                            } else {
                                this.pop();
                                this.raise();
                            }
                        }
                        break;
                }
                break;

            case OPCODES.FOR_ITER:
                switch (this.state) {
                    case 0:
                        if (this.top0().call('__next__')) {
                            return 1;
                        }
                    case 1:
                        if (!vm.return_value) {
                            this.pop();
                            if (except(MethodNotFoundError)) {
                                raise(TypeError, 'object does not support iteration');
                            } else if (except(StopIteration)) {
                                this.position = instruction.target;
                                break;
                            }
                        } else {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;


            case OPCODES.SETUP_LOOP:
            case OPCODES.SETUP_EXCEPT:
            case OPCODES.SETUP_FINALLY:
                this.push_block(OPCODES_EXTRA[instruction.opcode], instruction.target);
                break;

            case OPCODES.LOAD_CLOSURE:
                if (instruction.argument < this.code.cellvars.length) {
                    this.push(this.cells[this.code.cellvars[instruction.argument]]);
                } else {
                    this.push(this.cells[this.code.freevars[instruction.argument]]);
                }
                break;

            case OPCODES.LOAD_DEREF:
                if (instruction.argument < this.code.cellvars.length) {
                    this.push(this.cells[this.code.cellvars[instruction.argument]].get());
                } else {
                    this.push(this.cells[this.code.freevars[instruction.argument]].get());
                }
                break;

            case OPCODES.LOAD_CLASSDEREF:
                error('opcode not implemented');
                break;

            case OPCODES.STORE_DEREF:
                if (instruction.argument < this.code.cellvars.length) {
                    this.cells[this.code.cellvars[instruction.argument]].set(this.pop());
                } else {
                    error('load free variable closure not implemented');
                }
                break;

            case OPCODES.DELETE_DEREF:
                error('opcode not implemented');
                break;

            case OPCODES.RAISE_VARARGS:
                if (instruction.argument != 1) {
                    error('unsupported raise format');
                }
                exc_value = this.pop();
                raise(exc_value.cls, exc_value);
                break;

            case OPCODES.LOAD_CONST:
                this.push(this.code.constants[instruction.argument]);
                break;

            case OPCODES.CALL_FUNCTION:
                switch (this.state) {
                    case 0:
                        low = instruction.argument & 0xFF;
                        high = instruction.argument >> 8;
                        kwargs = {};
                        for (index = 0; index < high; index++) {
                            value = this.pop();
                            kwargs[this.pop().value] = value;
                        }
                        args = this.popn(low);
                        if (call(this.pop(), args, kwargs)) {
                            return 1;
                        }
                    case 1:
                        if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.CALL_FUNCTION_VAR:
                switch (this.state) {
                    case 0:
                        if (call(unpack_sequence, [this.pop()])) {
                            return 1;
                        }
                    case 1:
                        if (!vm.return_value) {
                            break;
                        }
                        low = instruction.argument & 0xFF;
                        high = instruction.argument >> 8;
                        kwargs = {};
                        for (index = 0; index < high; index++) {
                            value = this.pop();
                            kwargs[this.pop().value] = value;
                        }
                        args = this.popn(low).concat(vm.return_value.array);
                        if (call(this.pop(), args, kwargs)) {
                            return 2;
                        }
                    case 3:
                        if (vm.return_value) {
                            this.push(vm.return_value);
                        }
                        break;
                }
                break;

            case OPCODES.CALL_FUNCTION_KW:
                error('opcode is not supported');
                break;

            case OPCODES.CALL_FUNCTION_VAR_KW:
                error('opcode is not supported');
                break;

            case OPCODES.MAKE_FUNCTION:
            case OPCODES.MAKE_CLOSURE:
                low = instruction.argument & 0xFF;
                mid = (instruction.argument >> 8) & 0xFF;
                high = (instruction.argument >> 16) & 0x7FFF;
                name = this.pop();
                code = this.pop();
                var options = {defaults: {}, annotations: {}, globals: this.globals};
                if (instruction.opcode == OPCODES.MAKE_CLOSURE) {
                    options.closure = this.pop();
                }
                if (high) {
                    temp = this.pop();
                    for (index = 0; index < high - 1; index++) {
                        options.annotations[unpack_str(temp.get(index))] = this.pop();
                    }
                }
                for (index = 0; index < mid; index++) {
                    value = this.pop();
                    options.defaults[this.pop().value] = value;
                }
                for (index = 0; index < low; index++) {
                    options.defaults[code.code.signature.argnames[index]] = this.pop();
                }
                this.push(new PyFunction(unpack_str(name), code.code, options));
                break;

            case OPCODES.BUILD_SLICE:
                this.push(new_slice.apply(null, this.popn(instruction.argument)));
                break;

            default:
                error('unknown opcode occoured ' + instruction.opcode);
                break;
        }
        this.state = 0;
    }
};


NativeFrame.prototype.execute = function () {
    
        if (threading.step()) {
            return;
        }
    

    
        if (debugging.step()) {
            return;
        }
    

    assert(!this.code.simple, 'native frames\'s code is simple');
    var result;
    try {
        result = this.code.func.apply(null, this.args.concat([this.state, this]));
    } catch (error) {
        if (!(error instanceof PyObject)) {
            error = pack_error(error);
        }
        raise(error.cls, error, undefined, true);
        vm.frame = this.back;
        
            debugging.trace_return(this);
        
        
            if (!vm.frame) {
                threading.finished();
            }
        
        return;
    }
    if (result == undefined || result instanceof PyObject) {
        if (vm.return_value) {
            if (result instanceof PyObject) {
                vm.return_value = result;
            } else {
                vm.return_value = None;
            }
        }
        vm.frame = this.back;
        
            debugging.trace_return(this);
        
        
            if (!vm.frame) {
                threading.finished();
            }
        
    } else {
        this.state = result;
    }
};

    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var vm = {
    frame: null,

    return_value: None,
    last_exception: null,

    simple_depth: 0
};


function suspend() {
    if (vm.simple_depth != 0) {
        raise(RuntimeError, 'unable to suspend interpreter within simple native frame');
    } else {
        
            threading.drop();
        
    }
}


function resume(object, args, kwargs) {
    if (vm.frame) {
        raise(RuntimeError, 'interpreter is already running');
    }

    if (object instanceof PyObject) {
        call(object, args, kwargs);
        
            vm.frame.thread = new Thread(vm.frame);
            vm.frame.thread.enqueue();
            vm.frame = null;
        
    } else if (object instanceof Frame) {
        
            object.thread.enqueue();
        
    } else {
        raise(TypeError, 'invalid type of object to resume to');
    }

    
        threading.resume();
    
}


function run() {
    var frame, state;

    while (vm.frame) {
        frame = vm.frame;
        state = frame.execute();
        if (state != undefined) {
            frame.set_state(state);
        }
    }

    
}


function main(module, argv) {
    if (vm.frame) {
        raise(RuntimeError, 'interpreter is already running');
    }
    if (!(module instanceof PythonModule)) {
        raise(TypeError, 'invalid type of main module');
    }

    
        if (threading.main.frame) {
            raise(RuntimeError, 'main thread is already running');
        }
    

    get_namespace('sys')['argv'] = new PyList((['<python>'].concat(argv || []).map(pack_str)));
    register_module('__main__', module);

    module.dict['__name__'] = pack_str('__main__');

    vm.frame = new PythonFrame(module.code, {
        builtins: builtins,
        locals: module.dict,
        globals: module.dict
    });

    
        vm.frame.thread = threading.main;
        vm.frame.thread.frame = vm.frame;

        vm.frame.thread.return_value = None;
        vm.frame.thread.last_exception = null;

        vm.frame.thread.enqueue();

        threading.resume();
    
}


function call(object, args, kwargs, defaults, closure, globals, namespace) {
    var code, result;

    while (true) {
        if (object instanceof Frame) {
            vm.frame = object;
            
                debugging.trace_call(vm.frame);
            
            return vm.frame;
        } else if (object instanceof PythonCode) {
            if ((object.flags & CODE_FLAGS.GENERATOR) != 0) {
                vm.return_value = new PyGenerator(object, new PythonFrame(object, {
                    back: null, defaults: defaults, args: args, kwargs: kwargs,
                    globals: globals, closure: closure, namespace: namespace
                }));
                return null;
            } else {
                vm.frame = new PythonFrame(object, {
                    back: vm.frame, defaults: defaults, args: args, kwargs: kwargs,
                    globals: globals, closure: closure, namespace: namespace
                });
                
                    debugging.trace_call(vm.frame);
                
                return vm.frame;
            }
        } else if (object instanceof NativeCode) {
            if (object.simple) {
                vm.simple_depth++;
                try {
                    result = object.func.apply(null, object.parse_args(args, kwargs, defaults));
                    vm.return_value = result || None;
                } catch (error) {
                    if (error instanceof PyObject) {
                        raise(error.cls, error, null, true);
                    } else {
                        raise(JSError, pack_error(error), null, true);
                    }
                } finally {
                    vm.simple_depth--;
                }
                return null;
            } else {
                vm.frame = new NativeFrame(object, {
                    back: vm.frame, defaults: defaults, args: args, kwargs: kwargs,
                    globals: globals, closure: closure, namespace: namespace
                });
                
                    debugging.trace_call(vm.frame);
                
                return vm.frame;
            }
        } else if (object instanceof PyFunction) {
            closure = object.closure;
            globals = object.globals;
            defaults = object.defaults;
            object = object.code;
        } else if (object instanceof PyMethod) {
            args = [object.self].concat(args);
            object = object.func;
        } else if (object instanceof PyObject) {
            result = object.call('__call__', args, kwargs);
            if (except(MethodNotFoundError)) {
                raise(TypeError, object.cls.name + ' object is not callable', null, true);
                return null;
            } else {
                return result;
            }
        } else if (object instanceof PythonModule) {
            vm.frame = new PythonFrame(object.code, {
                back: vm.frame, locals: object.dict, globals: object.dict
            });
            return vm.frame;
        } else {
            raise(TypeError, 'invalid low level callable \'' + object + '\'', null, true);
        }
    }
}


function raise(exc_type, exc_value, exc_tb, suppress) {
    var frame, next_tb;

    if (!vm.frame) {
        error(exc_value);
    }

    if (typeof exc_value == 'string') {
        exc_value = make_exception(exc_type, exc_value);
    }

    exc_value.context = vm.return_value === null ? vm.last_exception.exc_value : None;

    vm.return_value = null;

    if (!exc_tb) {
        frame = vm.frame;
        while (frame) {
            exc_tb = new PyTraceback(frame, frame.position - 1, frame.get_line_number(), exc_tb);
            frame = frame.back;
        }
    }

    exc_value.traceback = exc_tb;

    

    vm.last_exception = {exc_type: exc_type, exc_value: exc_value, exc_tb: exc_tb};

    
        debugging.trace_raise(exc_type, exc_value, exc_tb);
    

    if ((vm.frame instanceof NativeFrame || vm.simple_depth > 0) && !suppress) {
        throw exc_value;
    } else {
        return exc_value;
    }
}


function except(exc_type) {
    if (!vm.return_value && issubclass(vm.last_exception.exc_type, exc_type)) {
        vm.return_value = None;
        return true;
    }
    return false;
}


$.vm = vm;

$.suspend = suspend;
$.resume = resume;
$.main = main;
$.call = call;
$.raise = raise;
$.except = except;


    
        /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */





var threading = {
    counter: 0,
    limit: 5000,
    identifier: 1,
    thread: null,
    queue: [],
    resuming: false,
    main: null,
    registry: {},

    get_thread: function (identifier) {
        return threading.registry[identifier];
    },

    resume: function () {
        if (!threading.resumeing) {
            threading.resumeing = true;
            window.postMessage('jaspy-resume', '*');
        }
    },

    finished: function () {
        
        threading.thread.save();
        
            debugging.trace_thread_finished(threading.thread);
        
        threading.thread.finished = true;
        threading.thread.frame = null;
        delete threading.registry[threading.thread.identifier];
        if (threading.queue.length) {
            threading.thread = threading.queue.shift();
            threading.thread.restore();
            
            vm.frame = threading.thread.frame;
        } else {
            
            threading.counter = 0;
            threading.thread = null;
            vm.frame = null;
        }
    },

    step: function () {
        if (threading.counter > threading.limit) {
            threading.counter = 0;
            threading.resumeing = true;
            threading.thread.save();
            
            window.postMessage('jaspy-resume', '*');
            vm.frame = null;
            return true;
        }

        if (threading.thread.counter > threading.limit / threading.queue.length) {
            threading.thread.counter = 0;
            threading.thread.save();
            threading.thread.enqueue();
            
            threading.thread = threading.queue.shift();
            threading.thread.restore();
            
            vm.frame = threading.thread.frame;
            return true;
        }

        threading.counter++;
        threading.thread.counter++;
    },

    drop: function (requeue) {
        if (requeue) {
            threading.thread.enqueue();
        }
        threading.thread.save();
        if (threading.queue.length) {
            
            threading.thread = threading.queue.shift();
            threading.thread.restore();
            
            vm.frame = threading.thread.frame;
        } else {
            threading.counter = 0;
            threading.thread = null;
            vm.frame = null;
        }
    },

    wakeup: function (event) {
        if (event.source == window && event.data == 'jaspy-resume') {
            threading.resumeing = false;
            event.stopPropagation();
            if (!threading.thread) {
                threading.thread = threading.queue.shift();
            }
            
            threading.thread.restore();
            run();
        }
    }
};

var Thread = Class.extend({
    constructor: function Thread(frame, name) {
        this.frame = frame;
        this.counter = 0;
        this.finished = false;
        this.identifier = threading.identifier++;
        this.name = name || ('Thread-' + this.identifier);

        this.return_value = None;
        this.last_exception = null;

        
            this.debug_internal = false;

            if (debugging) {
                debugging.trace_thread_created(this);
            }
        

        threading.registry[this.identifier] = this;
    },

    toString: function () {
        return '<Thread ' + this.identifier.toString() + '>';
    },

    enqueue: function () {
        threading.queue.push(this);
    },

    save: function () {
        this.frame = vm.frame;

        this.return_value = vm.return_value;
        this.last_exception = vm.last_exception;
    },

    restore: function () {
        vm.frame = this.frame;

        vm.return_value = this.return_value;
        vm.last_exception = this.last_exception;
    },

    get_frame: function (number) {
        var frame = this.frame;
        while (number > 0) {
            frame = frame.back;
            number--;
        }
        return frame;
    }
});


var Lock = Class.extend({
    constructor: function (reentrant) {
        this.reentrant = reentrant;
        this.waiting = [];
        this.thread = null;
        this.acquired = 0;
    },

    acquire: function (thread) {
        thread = thread || threading.thread;
        if (!this.thread || (this.reentrant && this.thread === thread)) {
            this.thread = thread;
            this.acquired++;
            return true;
        }
        this.waiting.push(thread);
        return false;
    },

    release: function (thread) {
        thread = thread || threading.thread;
        if (this.thread === thread) {
            this.acquired--;
            if (this.acquired == 0) {
                if (this.waiting.length) {
                    this.thread = this.waiting.shift();
                    this.thread.enqueue();
                } else {
                    this.thread = null;
                }
                return true;
            }
            return false;
        }
        raise(ValueError, 'unable to release lock, lock is not acquired by given thread');
    },

    remove: function (thread) {
        thread = thread || threading.thread;
        var index = this.waiting.indexOf(thread);
        if (index > -1) {
            this.waiting.splice(index, 1);
        } else {
            raise(ValueError, 'unable to remove thread from waiting threads, thread not found');
        }
    },

    locked: function () {
        return this.thread != null;
    }
});


window.addEventListener('message', threading.wakeup, true);


threading.main = new Thread(null, 'MainThread');


$.threading = threading;

$.Thread = Thread;
$.Lock = Lock;

    

    
        /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */


var Future = Class.extend({
    constructor: function () {
        this.result = null;
        this.exception = null;

        this.success = false;
        this.error = false;

        this.callbacks = [];
    },

    done: function (callback) {
        this.callbacks.push(callback);
    },

    run_callbacks: function () {
        for (var index = 0; index < this.callbacks.length; index++) {
            this.callbacks[index](this)
        }
    },

    set_result: function (result) {
        this.result = result;
        this.success = true;
        this.run_callbacks();
    },

    set_exception: function (exception) {
        this.exception = exception;
        this.error = true;
        this.run_callbacks();
    }
});



var debug_run_in_frame = new NativeCode(function (seq, in_frame, code, state, frame) {
    var py_frame;
    switch (state) {
        case 0:
            py_frame = new PythonFrame(code, {
                back: frame, locals: in_frame.locals, globals: in_frame.globals,
                builtins: in_frame.builtins
            });
            if (call(py_frame)) {
                return 1;
            }
        case 1:
            if (!vm.return_value) {
                // TODO: send exception data back
                return;
            }
            if (vm.return_value.call('__repr__')) {
                return 2;
            }
        case 2:
            if (!vm.return_value) {
                // TODO: send exception data back
                return;
            }
            debugging.emit('success', [vm.return_value.toString()], seq);
            return;
    }
}, {name: 'run_in_frame'}, ['seq', 'in_frame', 'code']);


var debug_get_frame_locals = new NativeCode(function (in_frame, state, frame) {
    var name;
    while (true) {
        switch (state) {
            case 0:
                frame.names = Object.keys(in_frame.locals);
                frame.result = {};
                frame.index = 0;
            case 1:
                if (frame.index >= frame.names.length) {
                    state = 3;
                    break;
                }
                if (in_frame.locals[frame.names[frame.index]].call('__repr__')) {
                    return 2;
                }
            case 2:
                // TODO: handle exceptions and deleted names
                if (!vm.return_value) {
                    console.log('error')
                } else {
                    name = frame.names[frame.index++];
                    frame.result[name] = {
                        type: in_frame.locals[name].cls.name,
                        value: vm.return_value.toString()
                    };
                }
                state = 1;
                break;
            case 3:
                return pack_object(frame.result);
        }
    }
}, {name: 'get_locals'}, ['in_frame']);

var debugging = {
    enabled: false,
    connected: false,

    module: null,
    argv: null,

    websocket: null,

    line_breakpoints: {},
    exception_breakpoints: {},

    console_patched: false,

    sequence_counter: 2,


    run: function () {
        main(debugging.module, debugging.argv);
        debugging.emit('running', [debugging.module.name, debugging.argv]);
    },

    add_break: function (filename, line, condition, expression) {
        if (!(filename in this.line_breakpoints)) {
            debugging.line_breakpoints[filename] = {};
        }
        debugging.line_breakpoints[filename][line] = {
            condition: condition, expression: expression
        };
        debugging.emit('break_added', [filename, line, condition, expression]);
    },

    remove_break: function (filename, line) {
        delete debugging.line_breakpoints[filename][line];
        debugging.emit('break_removed', [filename, line]);
    },

    add_exception_break: function (name, on_termination, on_raise) {
        debugging.exception_breakpoints[name] = {
            on_termination: on_termination, on_raise: on_raise
        };
        debugging.emit('exception_break_added', [name, on_termination, on_raise]);
    },

    remove_exception_break: function (name) {
        delete debugging.exception_breakpoints[name];
        debugging.emit('exception_break_removed', [name])
    },

    suspend_thread: function (thread_id) {
        threading.get_thread(thread_id).frame.debug_break = true;
    },

    resume_thread: function (thread_id) {
        var thread = threading.get_thread(thread_id);
        debugging.resume(thread);
    },

    kill_thread: function (thread_id) {
        var thread = threading.get_thread(thread_id);
        thread.restore();
        raise(SystemExit, 'thread has been killed by debugger', null, true);
        thread.save();
    },

    step_over: function (thread_id) {
        var thread = threading.get_thread(thread_id);
        if (!thread.debug_suspended) {
            throw new Error('Thread ' + thread_id + ' not suspended by debugger!');
        }
        thread.frame.debug_step_over = true;
        debugging.resume(thread);
    },

    step_into: function (thread_id) {
        var thread = threading.get_thread(thread_id);
        if (!thread.debug_suspended) {
            throw new Error('Thread ' + thread_id + ' not suspended by debugger!');
        }
        thread.frame.debug_step_into = true;
        debugging.resume(thread);
    },

    step_out: function (thread_id) {
        var thread = threading.get_thread(thread_id);
        if (!thread.debug_suspended) {
            throw new Error('Thread ' + thread_id + ' not suspended by debugger!');
        }
        thread.frame.debug_step_out = true;
        debugging.resume(thread);
    },

    call: function (native_code, args, kwargs) {
        var frame = new NativeFrame(native_code, {args: args, kwargs: kwargs});
        frame.debug_future = new Future();
        (new Thread(frame)).enqueue();
        threading.resume();
        return frame.debug_future;
    },

    connect: function (url) {
        debugging.websocket = new WebSocket(url);
        debugging.websocket.onopen = debugging.onopen;
        debugging.websocket.onerror = debugging.onerror;
        debugging.websocket.onmessage = debugging.onmessage;
        debugging.websocket.onclose = debugging.onclose;
    },

    onopen: function (event) {
        debugging.connected = true;
        debugging.emit_hello();
        debugging.emit_thread_created(threading.main);
    },

    onerror: function (event) {
        debugging.connected = false;
        alert('Connection to remote debugger has been aborted!');
    },

    onmessage: function (event) {
        var message = JSON.parse(event.data);
        try {
            debugging.commands[message.cmd].apply(null, [message.seq].concat(message.args));
        } catch (error) {
            debugging.emit_error(error.message, message.seq);
        }
    },

    onclose: function (event) {
        debugging.connected = false;
        alert('Connection to remote debugger has been closed!');
    },


    trace_line: function (frame, line) {
        if (frame.debug_step_over || frame.debug_step_into) {
            frame.debug_break = true;
            frame.debug_step_over = false;
            frame.debug_step_into = false;
        } else if (frame.code.filename in debugging.line_breakpoints) {
            // TODO: evaluate condition and expression
            if (debugging.line_breakpoints[frame.code.filename][line]) {
                frame.debug_break = true;
            }
        }
    },

    trace_call: function (frame) {
        if (frame.back && frame.back.debug_step_into) {
            frame.debug_break = true;
            frame.debug_step_into = false;
        }
    },

    trace_return: function (frame) {
        var return_value, last_exception;
        if (frame.debug_future) {
            if (vm.return_value) {
                frame.debug_future.set_result(vm.return_value);
            } else {
                frame.debug_future.set_exception(vm.last_exception);
            }
        }
        if (frame.back && (frame.debug_step_out || frame.debug_step_over)) {
            frame.back.debug_break = true;
            frame.debug_step_out = false;
            frame.debug_step_over = false;
        }
    },

    trace_raise: function (exc_type, exc_value, exc_tb) {
        // TODO: implement
    },

    trace_thread_created: function (thread) {
        if (!thread.frame.debug_internal) {
            thread.debug_intenal = true;
            debugging.emit_thread_created(thread);
        }
    },

    trace_thread_finished: function (thread) {
        if (!thread.debug_internal) {
            debugging.emit_thread_finished(thread);
        }
    },

    trace_console_log: function () {
        var index;
        var strings = [];
        for (index = 0; index < arguments.length; index++) {
            strings.push(arguments[index].toString());
        }
        debugging.emit('console_log', strings);
    },

    trace_console_error: function () {
        var index;
        var strings = [];
        for (index = 0; index < arguments.length; index++) {
            strings.push(arguments[index].toString());
        }
        debugging.emit('console_error', strings);
    },


    suspend: function () {
        var frame, frames;
        threading.thread.debug_suspended = true;
        frame = vm.frame;
        frames = [];
        while (frame) {
            frames.push({
                'file': frame.code.filename,
                'line': frame.get_line_number(),
                'name': frame.code.name
            });
            frame = frame.back;
        }
        debugging.emit('thread_suspended', [threading.thread.identifier, frames]);
        threading.drop();
    },

    resume: function (thread, first) {
        if (!thread.debug_suspended) {
            throw new Error('Unable to resume thread not suspended by debugger!');
        }
        thread.debug_suspended = false;
        thread.enqueue();
        threading.resume();
        debugging.emit('thread_resumed', [thread.identifier]);
    },

    step: function () {
        if (debugging.connected) {
            if (vm.frame.debug_break) {
                vm.frame.debug_break = false;
                debugging.suspend();
                return true;
            }
        }
    },

    emit: function (cmd, args, seq) {
        if (debugging.connected) {
            seq = seq == undefined ? debugging.sequence_counter += 2 : seq;
            debugging.websocket.send(JSON.stringify({'event': cmd, 'seq': seq, 'args': args || []}));
        }
    },

    emit_hello: function () {
        debugging.emit('hello');
    },

    emit_error: function (message, seq) {
        debugging.emit('error', [message], seq);
    },

    emit_thread_created: function (thread) {
        debugging.emit('thread_created', [thread.name, thread.identifier]);
    },

    emit_thread_finished: function (thread) {
        debugging.emit('thread_finished', [thread.identifier]);
    },



    commands: {
        run: function (seq) {
            if (threading.main.frame) {
                throw new Error('Main thread is already running!');
            }
            debugging.run();
        },

        resume_thread: function (seq, thread_id) {
            debugging.resume_thread(thread_id);
        },

        suspend_thread: function (seq, thread_id) {
            debugging.suspend_thread(thread_id);
        },

        kill_thread: function (seq, thread_id) {
            debugging.kill_thread(thread_id);
        },

        step_over: function (seq, thread_id) {
            debugging.step_over(thread_id);
        },

        step_into: function (seq, thread_id) {
            debugging.step_into(thread_id);
        },

        step_out: function (seq, thread_id) {
            debugging.step_out(thread_id);
        },

        run_in_frame: function (seq, thread_id, frame_id, source) {
            var frame = threading.get_thread(thread_id).get_frame(frame_id);
            debugging.call(debug_run_in_frame, [seq, frame, eval(source).code]);
        },

        run_in_thread: function (seq, source) {
            var code = eval(source);
            (new Thread(new PythonFrame(code.code))).enqueue();
            threading.resume();
        },

        js_eval: function (seq, source) {
            eval(source);
        },

        js_debugger: function (seq) {
            debugger;
        },

        get_threads: function (seq) {
            var identifier;
            var result = [];
            for (identifier in threading.registry) {
                if (threading.registry.hasOwnProperty(identifier)) {
                    result.push({
                        'id' : parseInt(identifier),
                        'name': threading.get_thread(identifier).name
                    });
                }
            }
            debugging.emit('threads', [result], seq);
        },

        get_locals: function (seq, thread_id, frame_id) {
            var frame = threading.get_thread(thread_id).get_frame(frame_id);
            debugging.call(debug_get_frame_locals, [frame]).done(function (future) {
                if (future.success) {
                    debugging.emit('locals', [future.result.primitive()], seq);
                } else {
                    debugging.emit_error('error receiving frame locals', seq)
                }
            });
        },

        add_break: function (seq, filename, line, condition, expression) {
            var condition_code = condition ? eval(condition).code : null;
            var expression_code = expression ? eval(expression).code : null;
            debugging.add_break(filename, line, condition_code, expression_code);
        },

        remove_break: function (seq, filename, line) {
            debugging.remove_break(filename, line);
        },

        add_exception_break: function (seq, name, on_termination, on_raise) {
            debugging.add_exception_break(name, on_termination, on_raise);
        },

        remove_exception_break: function (seq, name) {
            debugging.remove_exception_break(name);
        }
    }
};


function debugger_patch_console() {
    var log, error;

    function patched_log() {
        log.apply(window.console, arguments);
        disable_patch();
        try {
            debugging.trace_console_log.apply(null, arguments);
        } finally {
            enable_patch();
        }
    }

    function patched_error() {
        error.apply(window.console, arguments);
        disable_patch();
        try {
            debugging.trace_console_error.apply(null, arguments);
        } finally {
            enable_patch();
        }
    }

    function enable_patch() {
        window.console.log = patched_log;
        window.console.error = patched_error;
    }

    function disable_patch() {
        window.console.log = log;
        window.console.error = error;
    }
    
    debugging.console_patched = true;
    if (window.console) {
        log = window.console.log;
        error = window.console.error;
        enable_patch();
    } else {
        window.console = {};
        window.console.log = debugging.trace_console_log;
        window.console.error = debugging.trace_console_error;
    }
}


function debug(module, argv, url, options) {
    if (debugging.enabled) {
        raise(RuntimeError, 'debugger has already been started');
    }

    debugging.enabled = true;
    debugging.module = module;
    debugging.argv = argv;
    debugging.connect(url);

    options = options || {};

    if (options.patch_console && !debugging.console_patched) {
        debugger_patch_console();
    }
}


$.debugger = {};
$.debugger.run = debugging.run;
$.debugger.add_break = debugging.add_break;
$.debugger.remove_break = debugging.remove_break;
$.debugger.add_exception_break = debugging.add_exception_break;
$.debugger.remove_exception_break = debugging.remove_exception_break;
$.debugger.suspend_thread = debugging.suspend_thread;
$.debugger.resume_thread = debugging.resume_thread;
$.debugger.kill_thread = debugging.kill_thread;
$.debugger.step_over = debugging.step_over;
$.debugger.step_into = debugging.step_into;
$.debugger.step_out = debugging.step_out;
$.debugger.connect = debugging.connect;

$.debug = debug;

    

    return jaspy;
})();

/*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

jaspy.module('sys', function ($, module, builtins) {
    var VersionInfo = module.$class('_VersionInfo', [builtins.tuple]);

    VersionInfo.$def('__new__', function (cls, major, minor, micro, releaselevel, serial) {
        VersionInfo.check_subclass(cls);
        return new $.PyTuple([major, minor, micro, releaselevel, serial], cls);
    }, ['major', 'minor', 'patch', 'releaselevel', 'serial']);

    VersionInfo.$def('__str__', function (self) {
        VersionInfo.check(self);
        return $.pack_str('version_info(' +
            'major=' + self.array[0] + ', ' +
            'minor=' + self.array[1] + ', ' +
            'micro=' + self.array[2] + ', ' +
            'releaselevel=' + self.array[3].repr() + ', ' +
            'serial=' + self.array[4] + ')');
    });

    VersionInfo.$def_property('major', function (self) {
        VersionInfo.check(self);
        return self.array[0];
    });

    module.$set('byteorder', $.pack_str('big'));

    module.$set('copyright', $.pack_str('Copyright (C) 2016, Maximilian Koehl'));

    module.$set('platform', $.pack_str('web'));

    module.$set('implementation', $.pack_str('jaspy'));

    module.$set('maxunicode', $.pack_int(0xFFFF));

    module.$set('version', $.pack_str('3.5.1'));
    module.$set('version_info', new $.PyTuple([$.pack_int(3), $.pack_int(5), $.pack_int(0), $.pack_str('dev'), $.pack_str(0)], VersionInfo));

    module.$set('jaspy_version', $.pack_str('0.1.0.dev0'));
    module.$set('jaspy_version_info', new $.PyTuple([
        $.pack_int(0),
        $.pack_int(1),
        $.pack_int(0),
        $.pack_str('dev0'),
        $.pack_int(0)], VersionInfo)
    );

    module.$set('modules', new $.PyDict($.modules));
}, ['builtins']);


    /*
 * Copyright (C) 2016, Maximilian Koehl <mail@koehlma.de>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License version 3 as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along
 * with this program. If not, see <http://www.gnu.org/licenses/>.
 */

jaspy.module('_thread', function ($, module) {
    module.$def('start_new_thread', function (target, args, kwargs) {
        var frame = $.vm.frame;
        var new_frame = $.call(target, args, kwargs);
        new_frame.thread = new $.Thread(new_frame);
        new_frame.back = null;
        new_frame.thread.enqueue();
        $.vm.frame = frame;
    }, ['target', '*args', '**kwargs']);


    var LockType = module.$class('LockType');

    LockType.$def('__exit__', function (self, exc_type, exc_value, exc_tb) {
        LockType.check(self);
        self.lock.release();
        return $.False;
    }, ['exc_type', 'exc_value', 'exc_tb']);

    LockType.$def('acquire', function (self, state, frame) {
        LockType.check(self);
        switch (state) {
            case 0:
                if (self.lock.acquire()) {
                    return $.True;
                } else {
                    $.threading.drop();
                    return 1;
                }
            case 1:
                return $.True;
        }
    });

    LockType.$def('release', function (self) {
        LockType.check(self);
        self.lock.release();
    });

    LockType.$def('locked', function (self) {
        LockType.check(self);
        return self.lock.locked() ? $.True : $.False;
    });

    LockType.$def_alias('acquire', '__enter__');





    module.$def('allocate_lock', function () {
        var lock = LockType.make();
        lock.lock = new $.Lock();
        return lock;
    });



});




